# unit tests for deltaRCM_driver.py

import numpy as np

import pytest

import unittest.mock as mock

from pyDeltaRCM.model import DeltaModel
from pyDeltaRCM import shared_tools
from . import utilities


class Test__init__:

    def test_init(self, tmp_path):
        """
        test the deltaRCM_driver init (happened when delta.initialize was run)
        """
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        assert _delta.time_iter == 0.
        assert _delta._is_finalized is False
        # check that subclass parameter dictionary has been initialized
        assert _delta.subclass_parameters == {}

    def test_error_if_no_file_found(self):
        with pytest.raises(FileNotFoundError):
            _ = DeltaModel(input_file='./nonexisting_file.yaml')

    def test_override_single_default(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'S0', 0.005)
        f.close()
        delta = DeltaModel(input_file=p)
        assert delta.S0 == 0.005

    def test_override_two_defaults(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'S0', 0.005)
        utilities.write_parameter_to_file(f, 'Np_sed', 2)
        f.close()
        delta = DeltaModel(input_file=p)
        assert delta.S0 == 0.005
        assert delta.Np_sed == 2

    def test_override_single_default_kwarg(self, tmp_path):
        _out_dir = str(tmp_path / 'out_dir')
        delta = DeltaModel(out_dir=_out_dir, S0=0.005)
        assert delta.S0 == 0.005

    def test_override_two_defaults_kwargs(self, tmp_path):
        _out_dir = str(tmp_path / 'out_dir')
        delta = DeltaModel(
            out_dir=_out_dir,
            S0=0.005,
            Np_sed=2)
        assert delta.S0 == 0.005
        assert delta.Np_sed == 2

    def test_override_yaml_with_kwarg(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'S0', 0.005)
        f.close()
        with pytest.warns(UserWarning):
            delta = DeltaModel(input_file=p, S0=0.0333)
        assert delta.S0 == 0.0333

    def test_override_bad_type_float_string(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'S0', 'a string?!')
        f.close()
        with pytest.raises(TypeError):
            _ = DeltaModel(input_file=p)

    def test_override_bad_type_int_float(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'beta', 24.4234)
        f.close()
        with pytest.raises(TypeError):
            _ = DeltaModel(input_file=p)

    def test_not_creating_illegal_attributes(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'illegal_attribute', True)
        f.close()
        delta = DeltaModel(input_file=p)
        assert delta.S0 == 0.00015  # from default.yaml
        assert not hasattr(delta, 'illegal_attribute')

    def test_not_overwriting_existing_attributes(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'input_file', '/fake/path.yaml')
        f.close()
        delta = DeltaModel(input_file=p)
        assert delta.S0 == 0.00015  # from default.yaml
        assert hasattr(delta, 'input_file')
        assert delta.input_file == p

    def test_random_seed_settings_value(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'seed', 9999)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        f.close()
        shared_tools.set_random_seed(9999)
        _preval_same = shared_tools.get_random_uniform(1)
        shared_tools.set_random_seed(5)
        _preval_diff = shared_tools.get_random_uniform(1)
        delta = DeltaModel(input_file=p)
        assert delta.seed == 9999
        _postval_same = shared_tools.get_random_uniform(1)
        assert _preval_same == _postval_same
        assert delta.seed == 9999

    def test_random_seed_settings_newinteger_default(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'S0', 0.005)
        f.close()
        delta = DeltaModel(input_file=p)
        assert delta.seed is not None
        assert delta.seed <= (2**32) - 1
        assert isinstance(int(delta.seed), int)

    def test_negative_length(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'Length', -100)
        f.close()
        with pytest.raises(ValueError):
            _ = DeltaModel(input_file=p)

    def test_negative_width(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'Width', -100)
        f.close()
        with pytest.raises(ValueError):
            _ = DeltaModel(input_file=p)

    def test_negative_dx(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'dx', -100)
        f.close()
        with pytest.raises(ValueError):
            _ = DeltaModel(input_file=p)

    def test_bigger_than_Width_dx(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'Width', 10)
        utilities.write_parameter_to_file(f, 'dx', 100)
        f.close()
        with pytest.raises(ValueError):
            _ = DeltaModel(input_file=p)

    def test_bigger_than_Length_dx(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'Length', 10)
        utilities.write_parameter_to_file(f, 'dx', 100)
        f.close()
        with pytest.raises(ValueError):
            _ = DeltaModel(input_file=p)

    def test_negative_L0_meters(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'L0_meters', -100)
        f.close()
        with pytest.raises(ValueError):
            _ = DeltaModel(input_file=p)

    def test_negative_itermax(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'itermax', -100)
        f.close()
        with pytest.raises(ValueError):
            _ = DeltaModel(input_file=p)

    def test_negative_Np_water(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'Np_water', -100)
        f.close()
        with pytest.raises(ValueError):
            _ = DeltaModel(input_file=p)

    def test_negative_N0_meters(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'N0_meters', -100)
        f.close()
        with pytest.raises(ValueError):
            _ = DeltaModel(input_file=p)

    def test_negative_Np_sed(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'Np_sed', -100)
        f.close()
        with pytest.raises(ValueError):
            _ = DeltaModel(input_file=p)

    def test_negative_f_bedload(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'f_bedload', -100)
        f.close()
        with pytest.raises(ValueError):
            _ = DeltaModel(input_file=p)

    def test_big_f_bedload(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'f_bedload', 2)
        f.close()
        with pytest.raises(ValueError):
            _ = DeltaModel(input_file=p)

    def test_negative_C0_percent(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'C0_percent', -100)
        f.close()
        with pytest.raises(ValueError):
            _ = DeltaModel(input_file=p)

    def test_negative_Csmooth(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'Csmooth', -100)
        f.close()
        with pytest.raises(ValueError):
            _ = DeltaModel(input_file=p)

    def test_negative_active_layer_thickness(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'active_layer_thickness': -2.2})
        with pytest.raises(ValueError):
            _ = DeltaModel(input_file=p)

    def test_badtype_figure_saving(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _model = DeltaModel(input_file=p, defer_output=True)
        # add request to plot invalid attribute
        _model._save_fig_list['beta'] = ['beta']
        # try to finish init
        _model.init_output_file()
        with pytest.raises(AttributeError):
            _model.output_data()

    def test_badshape_figure_saving(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _model = DeltaModel(input_file=p, defer_output=True)
        # add request to plot attribute w/ invalid shape
        _model._save_fig_list['inlet'] = ['inlet']
        # try to finish init
        _model.init_output_file()
        with pytest.raises(AttributeError):
            _model.output_data()

    def test_clobber_netcdf(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'clobber_netcdf': True})
        _model = DeltaModel(input_file=p)
        # assert that model could have clobbered a netcdf
        assert _model._clobber_netcdf is True

    def test_clobbering(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'clobber_netcdf': True,
                                      'save_eta_grids': True})
        _model_1 = DeltaModel(input_file=p)
        _model_1.output_netcdf.close()
        # assert that model could have clobbered a netcdf
        assert _model_1._clobber_netcdf is True
        # make a second model which clobbers, raising eyebrows (and warning)
        with pytest.warns(UserWarning):
            _model_2 = DeltaModel(input_file=p)
        _model_2.output_netcdf.close()
        assert _model_2._clobber_netcdf is True

    def test_no_clobber_error(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml',
                                     {'save_eta_grids': True})
        _model_1 = DeltaModel(input_file=p)
        _model_1.output_netcdf.close()
        # assert that model could not have clobbered a netcdf
        assert _model_1._clobber_netcdf is False
        # make a second model which raises error
        with pytest.raises(FileExistsError):
            _ = DeltaModel(input_file=p)


class TestDeprecatedHooks:

    class HookDeltaModel(DeltaModel):
        """Dummy class to add old hook."""

        def hook_sed_route(self):
            """Old hook"""
            pass

    def test_if_hook_raise_error(self, tmp_path):
        """Deprecated hooks cannot be implemented and will raise an error."""

        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        with pytest.raises(AttributeError):
            _ = self.HookDeltaModel(input_file=p)

        # check that regular model still works though
        _ = DeltaModel(input_file=p)


class TestUpdate:

    def test_update(self, tmp_path):
        # create a delta with default settings
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # mock top-level methods, verify call was made to each
        _delta.solve_water_and_sediment_timestep = mock.MagicMock()
        _delta.apply_subsidence = mock.MagicMock()
        _delta.finalize_timestep = mock.MagicMock()
        _delta.log_model_time = mock.MagicMock()
        _delta.output_data = mock.MagicMock()
        _delta.output_checkpoint = mock.MagicMock()

        # run the timestep: t=0
        #   * should call all methods
        _delta.update()

        # assert calls
        assert _delta.solve_water_and_sediment_timestep.call_count == 1
        assert _delta.apply_subsidence.call_count == 1
        assert _delta.finalize_timestep.call_count == 1
        assert _delta.log_model_time.call_count == 1
        assert _delta.output_data.call_count == 1
        assert _delta.output_checkpoint.call_count == 1

        # assert times / counters
        assert _delta.time_iter == int(1)
        assert _delta.time == _delta.dt

        # run another step
        #   * should call all steps again
        _delta.update()

        # assert calls
        assert _delta.solve_water_and_sediment_timestep.call_count == 2
        assert _delta.apply_subsidence.call_count == 2
        assert _delta.finalize_timestep.call_count == 2
        assert _delta.log_model_time.call_count == 2
        assert _delta.output_data.call_count == 2
        assert _delta.output_checkpoint.call_count == 2

        # assert times / counters
        assert _delta.time_iter == int(2)
        assert _delta.time == 2 * _delta.dt

    def test_update_is_finalized(self, tmp_path):
        # create a delta with different itermax
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # change state to finalized
        _delta._is_finalized = True

        # run the timestep
        with pytest.raises(RuntimeError):
            _delta.update()


class TestFinalize:

    def test_finalize_not_updated(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        _delta.log_info = mock.MagicMock()
        _delta.output_data = mock.MagicMock()
        _delta.output_checkpoint = mock.MagicMock()

        # run finalize
        _delta.finalize()

        # assert calls
        #  should hit all options since no saves
        assert _delta.log_info.call_count == 2

        # these were originally included in `finalize`, but no longer.
        #   the checks for no call are here to ensure we don't revert
        assert _delta.output_data.call_count == 0
        assert _delta.output_checkpoint.call_count == 0

        assert _delta._is_finalized is True

    def test_finalize_is_finalized(self, tmp_path):
        # create a delta with different itermax
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # change state to finalized
        _delta._is_finalized = True

        # run the timestep
        with pytest.raises(RuntimeError):
            _delta.finalize()

        assert _delta._is_finalized is True


class TestPublicSettersAndGetters:

    def test_setting_getting_sea_surface_mean_elevation(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        assert _delta.sea_surface_mean_elevation == 0
        _delta.sea_surface_mean_elevation = 0.5
        assert _delta.sea_surface_mean_elevation == 0.5

    def test_setting_getting_sea_surface_elevation_change(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        assert _delta.sea_surface_elevation_change == 0
        _delta.sea_surface_elevation_change = 0.002
        assert _delta.sea_surface_elevation_change == 0.002

    def test_setting_getting_bedload_fraction(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        assert _delta.bedload_fraction == 0.5
        _delta.bedload_fraction = 0.25
        assert _delta.bedload_fraction == 0.25

    def test_setting_getting_channel_flow_velocity(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # mock the init methods
        _delta.create_boundary_conditions = mock.MagicMock()
        _delta.init_sediment_routers = mock.MagicMock()

        # check initials
        assert _delta.channel_flow_velocity == 1

        # change and assert changed
        _delta.channel_flow_velocity = 3
        assert _delta.channel_flow_velocity == 3
        assert _delta.channel_flow_velocity == _delta.u0
        assert _delta.channel_flow_velocity == _delta._u0

        # assert reinitializers called
        assert (_delta.create_boundary_conditions.called is True)
        assert (_delta.init_sediment_routers.called is True)

    def test_setting_getting_channel_width(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        _delta.create_boundary_conditions = mock.MagicMock()
        _delta.init_sediment_routers = mock.MagicMock()

        # check initials
        assert _delta.channel_width == 250

        # change value
        #  the channel width is then changed internally with `N0`, according
        #  to the `create_boundary_conditions` so no change is actually made
        #  here.
        with pytest.warns(UserWarning):
            _delta.channel_width = 300
        assert _delta.channel_width == 250  # not changed!

        # assert reinitializers called
        assert (_delta.create_boundary_conditions.called is True)
        assert (_delta.init_sediment_routers.called is True)

    def test_setting_getting_flow_depth(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # mock the init methods
        _delta.create_boundary_conditions = mock.MagicMock()
        _delta.init_sediment_routers = mock.MagicMock()

        # check initials
        assert _delta.channel_flow_depth == 5

        # change and assert changed
        _delta.channel_flow_depth = 2
        assert _delta.channel_flow_depth == 2
        assert _delta.channel_flow_depth == _delta.h0
        assert _delta.channel_flow_depth == _delta._h0

        # assert reinitializers called
        assert (_delta.create_boundary_conditions.called is True)
        assert (_delta.init_sediment_routers.called is True)

    def test_setting_getting_influx_concentration(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # mock the init methods
        _delta.create_boundary_conditions = mock.MagicMock()
        _delta.init_sediment_routers = mock.MagicMock()

        # check initials
        assert _delta.influx_sediment_concentration == 0.1

        # change and assert changed
        _delta.influx_sediment_concentration = 0.003
        assert _delta.C0_percent == 0.3

        # assert reinitializers called
        assert (_delta.create_boundary_conditions.called is True)
        assert (_delta.init_sediment_routers.called is True)

    def test_getter_nosetter_sea_surface_elevation(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # check that one is alias of other
        assert np.all(_delta.sea_surface_elevation == _delta.stage)
        assert (_delta.sea_surface_elevation is _delta.stage)

        # setter should return error, not allowed
        with pytest.raises(AttributeError):
            _delta.sea_surface_elevation = np.random.uniform(
                0, 1, size=_delta.eta.shape)

    def test_getter_nosetter_water_depth(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # check that one is alias of other
        assert np.all(_delta.water_depth == _delta.depth)
        assert (_delta.water_depth is _delta.depth)

        # setter should return error, not allowed
        with pytest.raises(AttributeError):
            _delta.water_depth = np.random.uniform(
                0, 1, size=_delta.eta.shape)

    def test_getter_nosetter_bed_elevation(self, tmp_path):
        p = utilities.yaml_from_dict(tmp_path, 'input.yaml')
        _delta = DeltaModel(input_file=p)

        # check that one is alias of other
        assert np.all(_delta.bed_elevation == _delta.eta)
        assert (_delta.bed_elevation is _delta.eta)

        # setter should return error, not allowed
        with pytest.raises(AttributeError):
            _delta.bed_elevation = np.random.uniform(
                0, 1, size=_delta.eta.shape)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

import pyDeltaRCM

# -- Project information -----------------------------------------------------

project = 'pyDeltaRCM'
copyright = '2020, The DeltaRCM Team'
author = 'The DeltaRCM Team'

# The full version, including alpha/beta/rc tags
release = pyDeltaRCM.__version__
version = pyDeltaRCM.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.doctest',
              'sphinx.ext.autosummary',
              'sphinx.ext.napoleon',
              'sphinx.ext.graphviz',
              'sphinx.ext.imgmath',
              'sphinx.ext.githubpages',
              'matplotlib.sphinxext.plot_directive',
              'sphinx.ext.todo']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# toggle todo items
todo_include_todos = True

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# Napoleon settings
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True

# Autosummary / Automodapi settings
autosummary_generate = True
automodapi_inheritance_diagram = False
autodoc_default_options = {'members': True, 'inherited-members': False,
                           'private-members': True}

# doctest
doctest_global_setup = '''
import pyDeltaRCM
import numpy as np
from matplotlib import pyplot as plt
'''
doctest_test_doctest_blocks = ''  # empty string disables testing all code in any docstring

## mpl plots
plot_basedir = 'pyplots'
plot_html_show_source_link = False
plot_formats = ['png', ('hires.png', 300)]
plot_pre_code = '''
import numpy as np
from matplotlib import pyplot as plt
import pyDeltaRCM
'''


# img math
# imgmath_latex_preamble = '\\usepackage{fouriernc}' # newtxsf, mathptmx

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinxdoc'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
# html_static_path = []




import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import mpl_toolkits.axes_grid1 as axtk
import abc

from . import shared_tools

# tools for water routing algorithms


class debug_tools(abc.ABC):
    """Debugging tools.

    These tools can be invoked as part of a script during runtime or in a
    Python debugging shell.

    Add ``breakpoint()`` to any line in the code to enter the debugger and use
    these tools interactively at that breakpoint. Note, for Python < 3.7 use
    ``pdb.set_trace()``.

    Examples
    --------

    Within a debugging shell:

    .. code::

        >>> self.show_attribute('cell_type', grid=False)
        >>> self.show_ind([3378, 9145, 11568, 514, 13558])
        >>> self.show_ind((42, 94), 'bs')
        >>> self.show_ind([(41, 8), (42, 10)], 'g^')
        >>> plt.show()

    .. plot:: debug_tools/debug_demo.py

    """
    def _get_attribute(self, attribute):
        _attr = getattr(self, attribute)
        if not isinstance(_attr, np.ndarray):
            raise TypeError('Attribute must be a numpy.ndarray, but was:'
                            '%s' % str(type(_attr)))
        _attr_shape = _attr.shape
        if len(_attr_shape) != 2:
            raise ValueError('Attribute "{at}" has shape {shp}, '
                             'but must be two-dimensional.'.format(at=attribute,
                                                                   shp=_attr_shape))
        return _attr

    def show_attribute(self, attribute, **kwargs):
        """Show an attribute over the model domain.

        Show any attribute of the :obj:`~pyDeltaRCM.model.DeltaModel` class
        over the model domain.

        Parameters
        ----------
        attribute : :obj:`str`
            Name of the attribute you want to show.

        ax : :obj:`matplotlib.Axes` object, optional
            Which axes to render attribute into. Uses ``gca()`` if no axis is
            provided.

        grid : :obj:`bool`, optional
            Whether to plot a grid over the domain to demarcate individual
            cells. Default is `True` (show the grid).

        block : :obj:`bool`, optional
            Whether to show the plot automatically. Default is `False` (do not
            show automatically).

        """

        if not isinstance(attribute, str):
            raise TypeError('Expected string for `attribute`, but was %s'
                            % type(attribute))
        _attr = self._get_attribute(attribute)
        plot_domain(_attr, **kwargs)

    def show_ind(self, ind, *args, **kwargs):
        """Show points within the model domain.

        Show the location of points (indices) within the model domain. Can
        show points as tuple ``(x, y)``, flat index ``idx``, list of tuples
        ``[(x1, y1), (x2, y2)]``, or list of flat indices ``[idx1, idx2]``.
        Method takes arbitrary `matplotlib` arguments to `plot` the points as
        Matlab-style args (``'r*'``) or keyword arguments (``marker='s'``).

        Parameters
        ----------
        ind : :obj:`tuple`, `int`, `list` of `tuple`, `list` of `int`, `list` of `int` and `tuple`
            Index (indices if list), to plot

        ax : :obj:`matplotlib.Axes` object, optional
            Which axes to render point into. Uses ``gca()`` if no axis is
            provided.

        block : :obj:`bool`, optional
            Whether to show the plot automatically. Default is `False` (do not
            show automatically).

        Other Parameters
        ----------------
        *args : :obj:`str`, optional
            Matlab-style point specifications, e.g., ``'r*'``, ``'bs'``.

        **kwargs : optional
            Any `kwargs` supported by `matplotlib.pyplot.plt`.

        """
        _shape = self.depth.shape
        if isinstance(ind, list) or isinstance(ind, np.ndarray):
            for _, iind in enumerate(ind):
                plot_ind(iind, shape=_shape, *args, **kwargs)
        else:
            plot_ind(ind, shape=_shape, *args, **kwargs)

    def show_line(self, ind, *args, multiline=False, nozeros=False, **kwargs):
        """Show line within the model domain.

        Show the location of lines (a series of connected indices) within the
        model domain. Can show points as a list/array of flat index ``[i0, i1,
        i2]``, list of tuples ``[(x1, y1), (x2, y2)]``, array of pairs ``[[x1,
        y1], [x2, y2]]``, or array of lines if given as flat index ``[[l0i0,
        l1i0, l2i0], [l0i1, l1i1, l2i1]]``. Method takes arbitrary
        `matplotlib` arguments to `plot` the points as Matlab-style args
        (``'r*'``) or keyword arguments (``marker='s'``).

        Parameters
        ----------
        ind : :obj: Nx2 `ndarray`, `list` of `tuple`, `list` of `int`, `ndarray` of `int`
            Indicies to plot.

        ax : :obj:`matplotlib.Axes` object, optional
            Which axes to render point into. Uses ``gca()`` if no axis is
            provided.

        block : :obj:`bool`, optional
            Whether to show the plot automatically. Default is `False` (do not
            show automatically).

        multiline : :obj:`bool`, optional
            When a 2D array is passed as `ind`, this boolean indicates whether
            to treat each column of the array as a separate line. Default is
            `False`.

        nozeros : :obj:`bool`, optional
            Whether to show segements of lines with a flat index `== 0 `.
            Default is `False`.

        Other Parameters
        ----------------
        *args : :obj:`str`, optional
            Matlab-style point specifications, e.g., ``'r*'``, ``'bs'``.

        **kwargs : optional
            Any `kwargs` supported by `matplotlib.pyplot.plt`.

        Returns
        -------
        lines : :obj:`list`
            Lines plotted by call to `show_line`, returned as a list, even if
            only one line plotted.

        """
        _shape = self.depth.shape
        if isinstance(ind, list):
            for _, iind in enumerate(ind):
                _l = plot_line(iind.flatten(),
                               shape=_shape, nozeros=nozeros,
                               *args, **kwargs)
                lines = [_l]
        elif isinstance(ind, np.ndarray):
            if ind.ndim > 2:
                raise NotImplementedError('Not implemented for arrays > 2d.')
            elif ind.ndim > 1:
                if multiline:
                    # travel along axis, extracting lines
                    cm = matplotlib.cm.get_cmap('tab10')
                    lines = []
                    for i in np.arange(ind.shape[1]):
                        _l = plot_line(ind[:, i], *args, multiline=multiline,
                                       nozeros=nozeros, shape=_shape,
                                       color=cm(i), **kwargs)
                        lines.append(_l)
                else:
                    _l = plot_line(ind, *args, nozeros=nozeros, **kwargs)
                    lines = [_l]
            else:
                _l = plot_line(ind.flatten(),
                               shape=_shape, nozeros=nozeros,
                               *args, **kwargs)
                lines = [_l]

        else:
            raise NotImplementedError
        return lines


def plot_domain(attr, ax=None, grid=True, block=False, label=None, **kwargs):
    """Plot the model domain.

    Public function to plot *any* 2d grid with helper display utils.
    This function is called by the DeltaModel method :obj:`show_attribute`.

    Parameters
    ----------
    attr : :obj:`ndarray`
        The data to display, as a 2D numpy array.

    ax : :obj:`matplotlib.Axes`, optional
        Which axes to render point into. Uses ``gca()`` if no axis is
            provided.

    grid : :obj:`bool`, optional
        Whether to plot a grid over the domain to demarcate individual
        cells. Default is `True` (show the grid).

    block : :obj:`bool`, optional
        Whether to show the plot automatically. Default is `False` (do not
        show automatically).

    label : :obj:`str`, optional
        A string describing the field being plotted. If given, will be
        appeneded to the colorbar of the domain plot.
    """
    attr_shape = attr.shape
    if not ax:
        ax = plt.gca()

    cmap = kwargs.pop('cmap', None)
    if (cmap is None):
        cmap = plt.get_cmap('viridis')
    # else:
        # cmap = plt.get_cmap(cmap)

    cobj = ax.imshow(attr,
                     cmap=cmap,
                     interpolation='none',
                     **kwargs)
    divider = axtk.axes_divider.make_axes_locatable(ax)
    cax = divider.append_axes("right", size="2%", pad=0.05)
    cbar = plt.colorbar(cobj, cax=cax)
    ax.autoscale(False)
    plt.sca(ax)

    if grid:
        shp = attr_shape
        ax.yaxis.set_major_locator(MaxNLocator(integer=True))
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        ax.set_xticks(np.arange(-.5, shp[1], 1), minor=True)
        ax.set_yticks(np.arange(-.5, shp[0], 1), minor=True)
        ax.tick_params(which='minor', length=0)
        ax.grid(which='minor', color='k', linestyle='-', linewidth=0.5)

    if label:
        cbar.ax.set_ylabel(label)

    if block:
        plt.show()


def plot_ind(_ind, *args, shape=None, **kwargs):
    """Plot points within the model domain.

    .. todo:: write a complete docstring with parameters etc.

    Method called by :obj:`show_ind`.

    Examples
    --------

    .. todo:: add examples, pull from tests.
    """
    ax = kwargs.pop('ax', None)
    block = kwargs.pop('block', False)

    _shape = shape

    if not ax:
        ax = plt.gca()

    if len(args) == 0:
        # args = 'r.',
        if not ('color' in kwargs.keys()):
            kwargs['color'] = 'red'
        if not ('marker' in kwargs.keys()):
            kwargs['marker'] = '.'
    if isinstance(_ind, tuple):
        if not len(_ind) == 2:
            raise ValueError('Expected tuple length to be 2, but was %s'
                             % str(len(_ind)))
    else:
        if (_shape is None):
            raise ValueError('Shape of array must be given to unravel index.')
        if isinstance(_ind, np.ndarray):
            _ind = _ind[0]
        _ind = shared_tools.custom_unravel(_ind, _shape)
    ax.plot(_ind[1], _ind[0], *args, **kwargs)

    if block:
        plt.show()


def plot_line(_ind, *args, shape=None, nozeros=False, **kwargs):
    """Plot a line within the model domain.

    Method called by :obj:`show_line`.

    .. todo:: write a complete docstring with parameters etc.

    Examples
    --------

    .. todo:: add examples, pull from tests.
    """
    ax = kwargs.pop('ax', None)
    block = kwargs.pop('block', False)
    multiline = kwargs.pop('multiline', False)

    _shape = shape

    if not ax:
        ax = plt.gca()

    if len(args) == 0:
        # args = 'k-',
        if not ('color' in kwargs.keys()):
            kwargs['color'] = 'k'
        if not ('marker' in kwargs.keys()):
            kwargs['ls'] = '-'
    if isinstance(_ind, tuple):
        raise NotImplementedError
        # if not len(_ind) == 2:
        #     raise ValueError('Expected tuple length to be 2, but was %s'
        #                      % str(len(_ind)))
    else:
        if isinstance(_ind, np.ndarray):
            if multiline:
                if (_shape is None):
                    raise ValueError(
                        'Shape of array must be given to unravel index.')
                pxpys = np.zeros((_ind.shape[0], 2))
                for i in range(_ind.shape[0]):
                    if _ind[i] == 0:
                        if nozeros:
                            pxpys[i, :] = np.nan, np.nan
                        else:
                            pxpys[i, :] = shared_tools.custom_unravel(_ind[i], _shape)
                    else:
                        pxpys[i, :] = shared_tools.custom_unravel(_ind[i], _shape)
            else:
                if _ind.ndim > 1:
                    pxpys = np.fliplr(_ind)
                else:
                    if (_shape is None):
                        raise ValueError(
                            'Shape of array must be given to unravel index.')
                    pxpys = np.zeros((_ind.shape[0], 2))
                    for i in range(_ind.shape[0]):
                        if _ind[i] == 0:
                            if nozeros:
                                pxpys[i, :] = np.nan, np.nan
                            else:
                                pxpys[i, :] = shared_tools.custom_unravel(_ind[i], _shape)
                        else:
                            pxpys[i, :] = shared_tools.custom_unravel(_ind[i], _shape)

    _l, = ax.plot(pxpys[:, 1], pxpys[:, 0], *args, **kwargs)

    if block:
        plt.show()

    return _l



import abc


class hook_tools(abc.ABC):
    """Tools defining various hooks in the model.

    For an explanation on model hooks, see the :doc:`User Guide
    </guides/user_guide>`.
    """

    def _check_deprecated_hooks(self):
        """Check for any old hooks that are defined.

        Sometimes hook names need to be deprecated, due to changing underlying
        model mechanics or anything else. This may mean that an old hook is
        *no longer called*, but there is no way to warn the user about this.
        Therefore, we enforce that no old hooks may be used as method names of
        subclassing models.

        This helper method is early in the `DeltaModel` `__init__` routine.
        """
        _deprecated_list = {'hook_sed_route': 'hook_route_sediment',
                            'hook_run_one_timestep': 'hook_solve_water_and_sediment_timestep'}
        for old_hook, new_hook in _deprecated_list.items():
            if hasattr(self, old_hook):
                raise AttributeError(
                    f'`{old_hook}` is deprecated, '
                    f'and has been replaced with `{new_hook}`.')

    def hook_import_files(self):
        """Hook :obj:`~pyDeltaRCM.init_tools.init_tools.import_files`.

        Called immediately before
        :obj:`~pyDeltaRCM.init_tools.init_tools.import_files`.

        This is the recommended hook to use for defining custom yaml parameters
        for a subclassed model. To allow the model to successfully load custom
        yaml parameters, expected types and default values need to be provided
        as dictionary key-value pairs to the model attribute,
        `subclass_parameters`.

        The expected format is:

        .. code::

            self.subclass_parameters['custom_param'] = {
                'type': ['expected_type'], 'default': default_value}  # noqa: E501

        """
        pass

    def hook_process_input_to_model(self):
        """Hook :obj:`~pyDeltaRCM.init_tools.init_tools.process_input_to_model`.

        Called immediately before
        :obj:`~pyDeltaRCM.init_tools.init_tools.process_input_to_model`.
        """
        pass

    def hook_create_other_variables(self):
        """Hook :obj:`~pyDeltaRCM.init_tools.init_tools.create_other_variables`.

        Called immediately before
        :obj:`~pyDeltaRCM.init_tools.init_tools.create_other_variables`.
        """
        pass

    def hook_create_domain(self):
        """Hook :obj:`~pyDeltaRCM.init_tools.init_tools.create_domain`.

        Called immediately before
        :obj:`~pyDeltaRCM.init_tools.init_tools.create_domain`.
        """
        pass

    def hook_load_checkpoint(self):
        """Hook :obj:`~pyDeltaRCM.init_tools.init_tools.load_checkpoint`.

        Called immediately before
        :obj:`~pyDeltaRCM.init_tools.init_tools.load_checkpoint`.
        """
        pass

    def hook_solve_water_and_sediment_timestep(self):
        """Hook :obj:`~pyDeltaRCM.iteration_tools.iteration_tools.solve_water_and_sediment_timestep`.

        Called immediately before
        :obj:`~pyDeltaRCM.iteration_tools.iteration_tools.solve_water_and_sediment_timestep`.
        """
        pass

    def hook_output_data(self):
        """Hook :obj:`~pyDeltaRCM.iteration_tools.iteration_tools.output_data`.

        Called immediately before
        :obj:`~pyDeltaRCM.iteration_tools.iteration_tools.output_data`.
        """
        pass

    def hook_apply_subsidence(self):
        """Hook :obj:`~pyDeltaRCM.iteration_tools.iteration_tools.apply_subsidence`.

        Called immediately before
        :obj:`~pyDeltaRCM.iteration_tools.iteration_tools.apply_subsidence`.
        """
        pass

    def hook_finalize_timestep(self):
        """Hook :obj:`~pyDeltaRCM.iteration_tools.iteration_tools.finalize_timestep`.

        Called immediately before
        :obj:`~pyDeltaRCM.iteration_tools.iteration_tools.finalize_timestep`.
        """
        pass

    def hook_output_checkpoint(self):
        """Hook :obj:`~pyDeltaRCM.iteration_tools.iteration_tools.output_checkpoint`.

        Called immediately before
        :obj:`~pyDeltaRCM.iteration_tools.iteration_tools.output_checkpoint`.
        """
        pass

    def hook_route_water(self):
        """Hook :obj:`~pyDeltaRCM.water_tools.water_tools.route_water`.

        Called immediately before
        :obj:`~pyDeltaRCM.iteration_tools.iteration_tools.route_water`.
        """
        pass

    def hook_init_water_iteration(self):
        """Hook :obj:`~pyDeltaRCM.water_tools.water_tools.init_water_iteration`.

        Called immediately before
        :obj:`~pyDeltaRCM.water_tools.water_tools.init_water_iteration`.
        """
        pass

    def hook_run_water_iteration(self):
        """Hook :obj:`~pyDeltaRCM.water_tools.water_tools.run_water_iteration`.

        Called immediately before
        :obj:`~pyDeltaRCM.water_tools.water_tools.run_water_iteration`.
        """
        pass

    def hook_compute_free_surface(self):
        """Hook :obj:`~pyDeltaRCM.water_tools.water_tools.compute_free_surface`.

        Called immediately before
        :obj:`~pyDeltaRCM.water_tools.water_tools.compute_free_surface`.
        """
        pass

    def hook_finalize_water_iteration(self, iteration):
        """Hook :obj:`~pyDeltaRCM.water_tools.water_tools.finalize_water_iteration`.

        Called immediately before
        :obj:`~pyDeltaRCM.water_tools.water_tools.finalize_water_iteration`.
        """
        pass

    def hook_route_sediment(self):
        """Hook :obj:`~pyDeltaRCM.sed_tools.sed_tools.route_sediment`.

        Called immediately before
        :obj:`~pyDeltaRCM.sed_tools.sed_tools.route_sediment`.
        """
        pass

    def hook_init_sediment_iteration(self):
        """Hook :obj:`~pyDeltaRCM.sed_tools.sed_tools.init_sediment_iteration`.

        Called immediately before
        :obj:`~pyDeltaRCM.sed_tools.sed_tools.init_sediment_iteration`.
        """
        pass

    def hook_route_all_sand_parcels(self):
        """Hook :obj:`~pyDeltaRCM.sed_tools.sed_tools.route_all_sand_parcels`.

        Called immediately before
        :obj:`~pyDeltaRCM.sed_tools.sed_tools.route_all_sand_parcels`.
        """
        pass

    def hook_topo_diffusion(self):
        """Hook :obj:`~pyDeltaRCM.sed_tools.sed_tools.topo_diffusion`.

        Called immediately before
        :obj:`~pyDeltaRCM.sed_tools.sed_tools.topo_diffusion`.
        """
        pass

    def hook_route_all_mud_parcels(self):
        """Hook :obj:`~pyDeltaRCM.sed_tools.sed_tools.route_all_mud_parcels`.

        Called immediately before
        :obj:`~pyDeltaRCM.sed_tools.sed_tools.route_all_mud_parcels`.
        """
        pass

    def hook_compute_sand_frac(self):
        """Hook :obj:`~pyDeltaRCM.iteration_tools.iteration_tools.compute_sand_frac`.

        Called immediately before
        :obj:`~pyDeltaRCM.iteration_tools.iteration_tools.compute_sand_frac`.
        """
        pass

    def hook_init_output_file(self):
        """Hook :obj:`~pyDeltaRCM.init_tools.init_tools.init_output_file`.

        Called immediately before
        :obj:`~pyDeltaRCM.init_tools.init_tools.init_output_file`.

        Expected format for entries to the `meta` dictionary nested within the
        `self._save_var_list` dictionary:

        .. code::

            self._save_var_list['meta']['new_meta_name'] = ['varname', 'units',
                                                            'type', (dimensions)]  # noqa: E501

        Expected format for time-varying metadata entries to the `meta`
        dictionary nested within the `self._save_var_list` dictionary:

        .. code::

            self._save_var_list['meta']['new_meta_attribute_name'] = [
                None, 'units', 'type', (dimensions)]  # noqa: E501

        .. note::

            For a vector of time-varying metadata, the dimension
            should be specified as ('total_time').

        Expected format for time varying grid entries as keys within the
        `self._save_var_list` dictionary:

        .. code::

            self._save_var_list['new_grid_name'] = ['varname', 'units',
                                                    'type', (dimensions)]

        """
        pass

    def hook_after_create_domain(self):
        """Hook called *after* :obj:`~pyDeltaRCM.init_tools.init_tools.create_domain`.

        Unlike the standard model hooks, this hook is called *after* the
        domain has been created with the
        method :obj:`~pyDeltaRCM.init_tools.init_tools.create_domain`
        """
        pass

    def hook_after_route_water(self):
        """Hook called *after* :obj:`~pyDeltaRCM.water_tools.water_tools.route_water`.

        Unlike the standard model hooks, this hook is called *after* water
        routing has completed with the
        method :obj:`~pyDeltaRCM.water_tools.water_tools.route_water`
        """
        pass

    def hook_after_route_sediment(self):
        """Hook called *after* :obj:`~pyDeltaRCM.sed_tools.sed_tools.route_sediment`.

        Unlike the standard model hooks, this hook is called *after* sed
        routing has completed with the
        method :obj:`~pyDeltaRCM.sed_tools.sed_tools.route_sediment`
        """
        pass

    def hook_after_finalize_timestep(self):
        """Hook called *after* :obj:`~pyDeltaRCM.iteration_tools.finalize_timestep`.

        Unlike the standard model hooks, this hook is called *after* the
        finalize timestep method has completed. 

        .. hint::

            This is the last method called before logging and outputting data
            for each timestep.
        """
        pass



import warnings

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

import pyDeltaRCM


# filter out the warning raised about no netcdf being found
warnings.filterwarnings("ignore", category=UserWarning)


n = 10
cm = matplotlib.cm.get_cmap('tab10')


# init delta model
with pyDeltaRCM.shared_tools._docs_temp_directory() as output_dir:
    delta = pyDeltaRCM.DeltaModel(
        out_dir=output_dir)

    delta_later = pyDeltaRCM.DeltaModel(
        out_dir=output_dir,
        resume_checkpoint='../../_resources/checkpoint')


_shp = delta_later.eta.shape


# manually call only the necessary paths
delta.init_water_iteration()
delta.run_water_iteration()

delta_later.init_water_iteration()
delta_later.run_water_iteration()


# define a function to fill in the walks of given idx
def _plot_idxs_walks_to_step(delta_inds, _step, _idxs, _ax):
    for i in range(len(_idxs)):
        iidx = _idxs[i]
        walk = delta_inds[iidx, :]
        walk = walk[:_step]
        pyDeltaRCM.debug_tools.plot_line(
            walk, shape=_shp, color=cm(i),
            nozeros=True)
        yend, xend = pyDeltaRCM.shared_tools.custom_unravel(
            walk[-1], _shp)
        _ax.plot(xend, yend,
                 marker='o', ms=3, color=cm(i))


# declare the idxs to use:
idxs = np.random.randint(low=0, high=delta._Np_water, size=n)
ps = [5, 25, 75]

# set up axis
fig, ax = plt.subplots(2, 3, sharex=True, sharey=True, figsize=(12, 4))
vmin, vmax = delta.eta.min(), delta.eta.max()

# fill in axis0
pyDeltaRCM.debug_tools.plot_domain(
    delta.eta, ax=ax[0, 0], grid=False, cmap='cividis')
_plot_idxs_walks_to_step(
    delta.free_surf_walk_inds, _step=ps[0], _idxs=idxs, _ax=ax[0, 0])
ax[0, 0].set_title('after {} steps'.format(ps[0]))
pyDeltaRCM.debug_tools.plot_domain(
    delta_later.eta, ax=ax[1, 0], grid=False, vmin=vmin, vmax=vmax, cmap='cividis')
_plot_idxs_walks_to_step(
    delta_later.free_surf_walk_inds, _step=ps[0], _idxs=idxs, _ax=ax[1, 0])
# ax[1, 0].set_title('after {} steps'.format(ps[0]))


# fill in axis1
pyDeltaRCM.debug_tools.plot_domain(
    delta.eta, ax=ax[0, 1], grid=False, cmap='cividis')
_plot_idxs_walks_to_step(
    delta.free_surf_walk_inds, _step=ps[1], _idxs=idxs, _ax=ax[0, 1])
ax[0, 1].set_title('after {} steps'.format(ps[1]))
pyDeltaRCM.debug_tools.plot_domain(
    delta_later.eta, ax=ax[1, 1], grid=False, vmin=vmin, vmax=vmax, cmap='cividis')
_plot_idxs_walks_to_step(
    delta_later.free_surf_walk_inds, _step=ps[1], _idxs=idxs, _ax=ax[1, 1])
# ax[1, 1].set_title('after {} steps'.format(ps[1]))


# fill in axis2
pyDeltaRCM.debug_tools.plot_domain(
    delta.eta, ax=ax[0, 2], grid=False, cmap='cividis')
_plot_idxs_walks_to_step(
    delta.free_surf_walk_inds, _step=ps[2], _idxs=idxs, _ax=ax[0, 2])
ax[0, 2].set_title('after {} steps'.format(ps[2]))
pyDeltaRCM.debug_tools.plot_domain(
    delta_later.eta, ax=ax[1, 2], grid=False, vmin=vmin, vmax=vmax, cmap='cividis')
_plot_idxs_walks_to_step(
    delta_later.free_surf_walk_inds, _step=ps[2], _idxs=idxs, _ax=ax[1, 2])
# ax[1, 3].set_title('after {} steps'.format(ps[3]))


plt.tight_layout()
plt.show()



import os
import sys
import glob

import unittest.mock as mock

from pyDeltaRCM.model import DeltaModel

from . import utilities


class TestLogModelTime:

    def test_records_arbitrary_time_values(self, tmp_path, capsys):
        """
        This test should create the log, and then print nothing at all.
        """
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        f.close()
        _delta = DeltaModel(input_file=p)

        _delta.logger = mock.MagicMock()

        # change values
        _delta._time = 3.14159
        _delta._time_iter = 42

        _delta.log_model_time()

        # will record whatever value is in _delta
        _calls = [mock.call('Time: 3.1; timestep: 42')]
        _delta.logger.info.assert_has_calls(_calls)

    def test_verbose_printing_0(self, tmp_path, capsys):
        """
        This test should create the log, and then print nothing at all.
        """
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'verbose', 0)
        f.close()
        _delta = DeltaModel(input_file=p)

        _delta.logger = mock.MagicMock()

        _delta.log_model_time()

        # time is logged regardless of verbosity
        _calls = [mock.call('Time: 0.0; timestep: 0')]
        _delta.logger.info.assert_has_calls(_calls)

        # stdout is empty because verbose 0
        captd = capsys.readouterr()
        assert captd.out == ''

    def test_verbose_printing_1(self, tmp_path, capsys):
        """
        This test should create the log, and then print nothing at all.
        """
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'verbose', 1)
        f.close()
        _delta = DeltaModel(input_file=p)

        _delta.logger = mock.MagicMock()

        _delta.log_model_time()

        # time is logged regardless of verbosity
        _calls = [mock.call('Time: 0.0; timestep: 0')]
        _delta.logger.info.assert_has_calls(_calls)

        # stdout has times because verbose 1
        captd = capsys.readouterr()
        assert 'Time: 0.0' in captd.out  # if verbose >= 1

    def test_verbose_printing_2(self, tmp_path, capsys):
        """
        This test should create the log, and then print nothing at all.
        """
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'verbose', 2)
        f.close()
        _delta = DeltaModel(input_file=p)

        _delta.logger = mock.MagicMock()

        _delta.log_model_time()

        # time is logged regardless of verbosity
        _calls = [mock.call('Time: 0.0; timestep: 0')]
        _delta.logger.info.assert_has_calls(_calls)

        # stdout is empty because verbose 2
        captd = capsys.readouterr()
        assert 'Time: 0.0' in captd.out  # if verbose >= 1


class TestLoggerIntegratedDuringInitialization:
    """Test that logger records during instantiation.

    These tests could be set up with a patch on the logger before
    instantiation of the DeltaModel...
    """

    def test_logger_has_initialization_lines(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'verbose', 1)
        utilities.write_parameter_to_file(f, 'seed', 10)
        f.close()
        _delta = DeltaModel(input_file=p)

        _logs = glob.glob(os.path.join(_delta.prefix, '*.log'))
        assert len(_logs) == 1  # log file exists
        with open(_logs[0], 'r') as _logfile:
            _lines = _logfile.readlines()
            _lines = ' '.join(_lines)  # collapse to a single string
            assert 'Setting model constants' in _lines
            assert 'Random seed is: 10' in _lines
            assert 'Creating model domain' in _lines
            assert 'Initializing output NetCDF4 file' in _lines
            assert 'Model initialization complete' in _lines

            if sys.platform.startswith('linux'):
                assert 'Platform: Linux-' in _lines
            elif sys.platform == 'darwin':
                guess1 = 'Platform: Darwin-' in _lines
                guess2 = 'Platform: macOS-' in _lines
                assert (guess1 | guess2)
            elif sys.platform.startswith('win'):
                assert 'Platform: Windows-' in _lines
            else:
                raise RuntimeError(
                    'Platform type not recognized.')
        assert not os.path.isfile(os.path.join(
            tmp_path, 'out_dir', 'discharge_0.0.png'))
        assert not os.path.isfile(os.path.join(
            tmp_path, 'out_dir', 'eta_0.0.png'))

    def test_logger_random_seed_always_recorded(self, tmp_path):
        file_name = 'user_parameters.yaml'
        p, f = utilities.create_temporary_file(tmp_path, file_name)
        utilities.write_parameter_to_file(f, 'out_dir', tmp_path / 'out_dir')
        utilities.write_parameter_to_file(f, 'verbose', 0)
        # do not set the seed explicitly, let it be set by the model
        f.close()
        delta = DeltaModel(input_file=p)
        _logs = glob.glob(os.path.join(delta.prefix, '*.log'))
        assert len(_logs) == 1  # log file exists
        with open(_logs[0], 'r') as _logfile:
            _lines = _logfile.readlines()
            _joinedlines = ' '.join(_lines)  # collapse to a single string
            assert 'Random seed is: ' in _joinedlines

            # determine the index of the line
            _idx = ['Random seed is: ' in _l for _l in _lines]
            assert sum(_idx) == 1  # one and only one True in list
            _idx = _idx.index(True)

            # try to covert to int, otherwise fail
            _seed = _lines[_idx].split(':')[-1]  # pull the seed value
            try:
                _intseed = int(_seed)
            except ValueError:
                raise ValueError('Could not convert the seed to int')

            assert _intseed >= 0




import numpy as np
from numba import njit
import abc

from . import shared_tools

# tools for water routing algorithms


class water_tools(abc.ABC):

    def route_water(self):
        """Water routing main method.

        This is the main method for water routing in the model. It is
        called once per `update()` call.

        Internally, this method calls:
            * :obj:`init_water_iteration`
            * :obj:`run_water_iteration`
            * :obj:`compute_free_surface`
            * :obj:`finalize_water_iteration`
        """
        _msg = 'Beginning water iteration'
        self.log_info(_msg, verbosity=2)

        for iteration in range(self._itermax):

            # initialize the relevant fields and parcel trackers
            self.hook_init_water_iteration()
            self.init_water_iteration()

            # run the actual iteration of the parcels
            self.hook_run_water_iteration()
            self.run_water_iteration()

            # accumulate the routed water parcels into free surface
            self.hook_compute_free_surface()
            self.compute_free_surface()

            # clean up the water surface and apply boundary conditions
            self.hook_finalize_water_iteration(iteration)
            self.finalize_water_iteration(iteration)

    def init_water_iteration(self):
        """Init the water iteration routine.
        """
        _msg = 'Initializing water iteration'
        self.log_info(_msg, verbosity=2)

        self.qxn[:] = 0
        self.qyn[:] = 0
        self.qwn[:] = 0

        self.free_surf_flag[:] = 1  # all parcels begin as valid
        self.free_surf_walk_inds[:] = 0
        self.sfc_visit[:] = 0
        self.sfc_sum[:] = 0

        self.pad_stage = np.pad(self.stage, 1, 'edge')
        self.pad_depth = np.pad(self.depth, 1, 'edge')
        self.pad_cell_type = np.pad(self.cell_type, 1, 'edge')

    def get_inlet_weights_water(self, **kwargs):
        """Get weight for inlet cells for water parcels.

        This method determines the *weights* describing which inlet cells
        water parcels are fed into the domain (where the cells are
        `self.inlet`).
        Internally, :obj:`pyDeltaRCM.shared_tools._get_inlet_weights` is
        called, returning a balanced weighting across all inlet cells.

        .. note::

            Reimplement this method in custom subclasses as needed.
        """
        return shared_tools.get_inlet_weights(self.inlet)

    def run_water_iteration(self):
        """Run a single iteration of travel paths for all water parcels.

        Runs all water  parcels (`Np_water` parcels) for `stepmax` steps, or
        until the parcels reach a boundary.

        All parcels are processed in parallel, taking one step for each loop
        of the ``while`` loop.

        Example
        -------
        On the initial delta surface, see how ten selected parcels are routed through the domain:

        .. plot:: water_tools/run_water_iteration.py
        """
        _msg = 'Beginning stepping of water parcels'
        self.log_info(_msg, verbosity=2)

        # configure the starting indices for each parcel
        inlet_weights = self.get_inlet_weights_water()
        start_indices = shared_tools.get_start_indices(self.inlet,
                                                       inlet_weights,
                                                       self._Np_water)

        # init parcel step number counter
        _step = 0

        # flux from ghost node
        self.qxn.flat[start_indices] += 1
        self.qyn.flat[start_indices] += 0  # this could be omitted...
        self.qwn.flat[start_indices] += self.Qp_water / self._dx / 2

        # load the initial indices into the walk indices
        self.free_surf_walk_inds[:, _step] = start_indices
        current_inds = np.copy(start_indices)

        # get weights, and flatten for faster access
        self.get_water_weight_array()
        water_weights_flat = self.water_weights.reshape(-1, 9)

        # while any parcles still need to take any steps
        while (sum(current_inds) > 0) & (_step < self.stepmax):

            # step counter increment
            _step += 1

            # check whether storage size needs to be increased
            self.check_size_of_indices_matrix(_step)

            # use water weights and random pick to determine d8 direction
            new_direction = _choose_next_directions(
                current_inds, water_weights_flat)

            # use the new directions for each parcel to determine the new ind
            #   for each parcel
            new_inds = _calculate_new_inds(
                current_inds,
                new_direction,
                self.ravel_walk_flat)

            # determine the step of each parcel made based on the new direction
            dist, istep, jstep, astep = shared_tools.get_steps(
                new_direction,
                self.iwalk_flat,
                self.jwalk_flat)

            # update the discharge field with walk of parcels
            #   this updates flux out of the old inds, and into the new inds
            self.update_Q(
                dist, current_inds, new_inds, astep, jstep, istep,
                update_current=True, update_next=True)

            # check for any loops in the walks of parcels
            #   A loop disqualifies a parcel from the free surface
            #   calculation, but does not stop the parcel from routing, and
            #   thus influencing the qw fields. If a parcel is looped, it
            #   will be given a new location in the domain, along the
            #   mean-transport vector. See function for full description.
            new_inds, looped = _check_for_loops(
                self.free_surf_walk_inds, new_inds, _step, self.L0,
                self.CTR, self.stage - self.H_SL)
            looped = looped.astype(bool)

            # set the current_inds to be the new_inds values
            #   (i.e., take the step)
            current_inds[:] = new_inds[:]

            # invalidate the looped parcels from the free surface
            self.free_surf_flag[looped] = 0

            # check for parcels that have reached the boundary
            boundary = self.check_for_boundary(current_inds)

            # parcels that have looped and reached the boundary are set zero
            #   before saving their indices
            boundary_looped = np.logical_and(looped, boundary)
            current_inds[boundary_looped] = 0

            # Record the parcel pathways for computing the free surface later
            self.free_surf_walk_inds[:, _step] = current_inds

            # parcels that have reached the boundary are set ``ind==0``,
            #   effectively ending the routing of these parcels.
            current_inds[boundary] = 0

            # update the q*n fields for the final step
            #   to ensure flux balanced at domain edge
            if np.any(boundary):
                self.update_Q(
                    dist[boundary], current_inds[boundary],
                    new_inds[boundary], astep[boundary],
                    jstep[boundary], istep[boundary],
                    update_current=False, update_next=True)

    def compute_free_surface(self):
        """Calculate free surface after routing all water parcels.

        This method uses the `free_surf_walk_inds` matrix accumulated
        during the routing of the water parcels (in
        :obj:`run_water_iteration`) to determine the free surface. The
        operations of the free surface computation are placed in a jitted
        function :obj:`_accumulate_free_surface_walks`. Following this
        computation, the free surface is smoothed by steps in
        :obj:`finalize_free_surface`.

        See [1]_ and [2]_ for a complete description of hydrodynamic
        assumptions in the DeltaRCM model.

        Examples
        --------

        The sequence of `compute_free_surface` is depicted in the figures
        below. The first image depicts the "input" to `compute_free_surface`,
        which is the current bed elevation, and the path of each water parcel
        (cyan lines in right image).

        .. plot:: water_tools/compute_free_surface_inputs.py

        The `compute_free_surface` method then calls the
        :obj:`_accumulate_free_surface_walks` function to determine 1) the
        number of times each cell has been visited by a water parcel
        (``sfc_visit``), and 2) the *total sum of expected elevations* of the
        water surface at each cell (``sfc_sum``).

        The output from :obj:`_accumulate_free_surface_walks` is then used to
        calculate a new stage surface (``H_new``) based only on the water
        parcel paths and expected water surface elevations, approximately as
        ``Hnew = sfc_sum / sfc_visit`` ("computed Hnew" in figure below)

        Following this step, a correction is applied forcing the new free
        surface to be below sea level and above or equal to the land surface
        elevation over the model domain ("stage" in figure below). This
        surface `Hnew` is used in the following operation
        :obj:`finalize_free_surface`.

        Finally, the updated water surface is
        combined with the previous timestep's water surface and an
        underrelaxation coefficient
        :obj:`~pyDeltaRCM.model.DeltaModel.omega_sfc`.

        .. plot:: water_tools/compute_free_surface_outputs.py

        .. [1] A reduced-complexity model for river delta formation – Part 1:
               Modeling deltas with channel dynamics, M. Liang, V. R. Voller,
               and C. Paola, Earth Surf. Dynam., 3, 67–86, 2015.
               https://doi.org/10.5194/esurf-3-67-2015

        .. [2] A reduced-complexity model for river delta formation – Part 2:
               Assessment of the flow routing scheme, M. Liang, N. Geleynse,
               D. A. Edmonds, and P. Passalacqua, Earth Surf. Dynam., 3,
               87–104, 2015. https://doi.org/10.5194/esurf-3-87-2015

        """
        _msg = 'Computing free surface from water parcels'
        self.log_info(_msg, verbosity=2)

        self.sfc_visit, self.sfc_sum = _accumulate_free_surface_walks(
            self.free_surf_walk_inds, self.free_surf_flag, self.cell_type,
            self.uw, self.ux, self.uy, self.depth,
            self._dx, self._u0, self.h0, self._H_SL, self._S0)

        # begin from the previous stage
        Hnew = self.eta + self.depth

        # find average water surface elevation for a cell from accumulation
        Hnew[self.sfc_visit > 0] = (self.sfc_sum[self.sfc_visit > 0] /
                                    self.sfc_visit[self.sfc_visit > 0])

        # water surface height not under sea level
        Hnew[Hnew < self._H_SL] = self._H_SL

        # water surface height not below bed elevation
        Hnew[Hnew < self.eta] = self.eta[Hnew < self.eta]

        # set to model field
        self.Hnew = Hnew

        # finalize the free surface (combine and smooth)
        self.finalize_free_surface()

    def finalize_free_surface(self):
        """Finalize the water free surface.

        This method occurs after the initial computation of the free surface,
        and creates the new free surface by smoothing the newly computed free
        surface with a jitted function (:obj:`_smooth_free_surface`), and then
        combining the old surface and new smoothed surface with
        underrelaxation. Finally, a :obj:`flooding_correction` is applied.
        """
        _msg = 'Smoothing and finalizing free surface'
        self.log_info(_msg, verbosity=2)

        # smooth newly calculated free surface
        Hsmth = _smooth_free_surface(
            self.Hnew, self.cell_type, self._Nsmooth, self._Csmooth)

        # combine new smoothed and previous free surf with underrelaxation
        if self._time_iter > 0:
            self.stage = (((1 - self._omega_sfc) * self.stage) +
                          (self._omega_sfc * Hsmth))

        # apply a flooding correction
        self.flooding_correction()

    def finalize_water_iteration(self, iteration):
        """Finish updating flow fields.

        Clean up at end of water iteration
        """
        _msg = 'Finalizing stepping of water parcels'
        self.log_info(_msg, verbosity=2)

        # apply boundary for water surface
        #     Note: not in Matlab implementation
        self.stage[:] = np.maximum(self.stage, self._H_SL)

        # apply an update on the depth to match stage and eta values
        self.depth[:] = np.maximum(self.stage - self.eta, 0)  # never negative

        # update fields for sediment routing
        self.update_flow_field(iteration)
        self.update_velocity_field()

    def check_size_of_indices_matrix(self, it):
        """Check if step path matrix needs to be made larger.

        Initial size of self.free_surf_walk_inds is half of self.stepmax
        because the number of iterations doesn't go beyond
        that for many timesteps.

        Once it reaches it > self.stepmax/2 once, make the size
        self.iter for all further timesteps
        """
        if it >= self.free_surf_walk_inds.shape[1]:
            _msg = 'Increasing size of self.free_surf_walk_inds'
            self.log_info(_msg, verbosity=2)

            indices_blank = np.zeros(
                (int(self._Np_water), int(self.stepmax / 4)), dtype=int)

            self.free_surf_walk_inds = np.hstack((self.free_surf_walk_inds, indices_blank))

    def get_water_weight_array(self):
        """Get step direction weights for each cell.

        This method is called once, before parcels are stepped, because the
        weights do not change during the stepping of parcels.

        See :doc:`/info/hydrodynamics` for a description of the model design
        and equations/algorithms for how water weights are determined.

        .. note::

            No computation actually occurs in this method. Internally, the
            method calls the jitted :obj:`_get_water_weight_array`, which in
            turn loops through all of the cells in the model domain and
            determines the water weight for that cell with
            :obj:`get_weight_sfc_int` and :obj:`_get_weight_at_cell_water`.

        Examples
        --------

        The following figure shows several examples of locations within the
        model domain, and the corresponding water routing weights determined
        for that location.

        .. plot:: water_tools/water_weights_examples.py
        """
        _msg = 'Computing water weight array'
        self.log_info(_msg, verbosity=2)

        # compiling the water weight array is handled inside a jitted
        #     function below. ~4x faster.
        self.water_weights = _get_water_weight_array(
            self.depth, self.stage, self.cell_type, self.qx, self.qy,
            self.ivec_flat, self.jvec_flat, self.distances_flat,
            self.dry_depth, self.gamma, self._theta_water)

    def update_Q(self, dist, current_inds, next_inds, astep, jstep, istep,
                 update_current=False, update_next=False):
        """Update discharge field values.

        Method is called after one set of water parcel steps.
        """
        _msg = 'Updating flux fields after single parcel step'
        self.log_info(_msg, verbosity=2)

        if update_current:
            self.qxn = _update_dirQfield(
                self.qxn.flat[:], dist, current_inds,
                astep, jstep).reshape(self.qxn.shape)
            self.qyn = _update_dirQfield(
                self.qyn.flat[:], dist, current_inds,
                astep, istep).reshape(self.qyn.shape)
            self.qwn = _update_absQfield(
                self.qwn.flat[:], dist, current_inds,
                astep, self.Qp_water, self._dx).reshape(self.qwn.shape)
        if update_next:
            self.qxn = _update_dirQfield(
                self.qxn.flat[:], dist, next_inds,
                astep, jstep).reshape(self.qxn.shape)
            self.qyn = _update_dirQfield(
                self.qyn.flat[:], dist, next_inds,
                astep, istep).reshape(self.qyn.shape)
            self.qwn = _update_absQfield(
                self.qwn.flat[:], dist, next_inds,
                astep, self.Qp_water, self._dx).reshape(self.qwn.shape)

    def check_for_boundary(self, inds):
        """Check whether parcels have reached the boundary.

        Checks whether any parcels have reached the model boundaries. If they
        have, then update the information in
        :obj:`~pyDeltaRCM.DeltaModel.free_surf_flag`.

        Parameters
        ----------
        inds : :obj:`ndarray`
            Unraveled indicies of parcels.
        """
        _msg = 'Checking stepped parcels against boundary location'
        self.log_info(_msg, verbosity=2)

        # inds[self.free_surf_flag == 2] = 0
        boundary = (self.cell_type.flat[inds] == -1)
        return boundary

    def update_flow_field(self, iteration):
        """Update water discharge.

        Update the water discharge field after one set of water parcels
        iteration.
        """
        _msg = 'Updating discharge fields after parcel stepping'
        self.log_info(_msg, verbosity=2)

        dloc = (self.qxn**2 + self.qyn**2)**(0.5)

        qwn_div = np.ones((self.L, self.W))
        qwn_div[dloc > 0] = self.qwn[dloc > 0] / dloc[dloc > 0]

        self.qxn *= qwn_div
        self.qyn *= qwn_div

        if self._time_iter > 0:

            omega = self.omega_flow_iter
            if iteration == 0:
                omega = self._omega_flow

            self.qx = self.qxn * omega + self.qx * (1 - omega)
            self.qy = self.qyn * omega + self.qy * (1 - omega)

        else:
            self.qx = self.qxn.copy()
            self.qy = self.qyn.copy()

        self.qw = (self.qx**2 + self.qy**2)**(0.5)

        self.qx[0, self.inlet] = self.qw0
        self.qy[0, self.inlet] = 0
        self.qw[0, self.inlet] = self.qw0

    def update_velocity_field(self):
        """Update flow velocity fields.

        Update the flow velocity fields after one set of water parcels
        iteration.
        """
        _msg = 'Updating flow velocity fields after parcel stepping'
        self.log_info(_msg, verbosity=2)

        mask = np.logical_and((self.depth > self.dry_depth), (self.qw > 0))

        self.uw[mask] = np.minimum(self.u_max,
                                   self.qw[mask] / self.depth[mask])
        self.ux[mask] = self.uw[mask] * self.qx[mask] / self.qw[mask]
        self.uy[mask] = self.uw[mask] * self.qy[mask] / self.qw[mask]

        self.uw[~mask] = 0
        self.ux[~mask] = 0
        self.uy[~mask] = 0

    def flooding_correction(self):
        """Flood dry cells along the shore if necessary.

        Check the neighbors of all dry cells. If any dry cells have wet
        neighbors, check that their stage is not higher than the bed elevation
        of the center cell.
        If it is, flood the dry cell.
        """
        _msg = 'Computing flooding correction'
        self.log_info(_msg, verbosity=2)

        wet_mask = self.depth > self.dry_depth
        wet_mask_nh = self.get_wet_mask_nh()
        wet_mask_nh_sum = np.sum(wet_mask_nh, axis=0)

        # makes wet cells look like they have only dry neighbors
        wet_mask_nh_sum[wet_mask] = 0

        # indices of dry cells with wet neighbors
        shore_ind = np.where(wet_mask_nh_sum > 0)

        stage_nhs = self.build_weight_array(self.stage)
        eta_shore = self.eta[shore_ind]

        for i in range(len(shore_ind[0])):

            # pretends dry neighbor cells have stage zero
            #    so they cannot be > eta_shore[i]
            stage_nh = wet_mask_nh[:, shore_ind[0][i], shore_ind[1][i]] * \
                stage_nhs[:, shore_ind[0][i], shore_ind[1][i]]

            if (stage_nh > eta_shore[i]).any():
                self.stage[shore_ind[0][i], shore_ind[1][i]] = max(stage_nh)

    def get_wet_mask_nh(self):
        """Get wet mask.

        Returns np.array((8,L,W)), for each neighbor around a cell
        with 1 if the neighbor is wet and 0 if dry
        """
        wet_mask = (self.depth > self.dry_depth) * 1
        wet_mask_nh = self.build_weight_array(wet_mask, fix_edges=True)

        return wet_mask_nh

    def build_weight_array(self, array, fix_edges=False, normalize=False):
        """Weighting array of neighbors.

        Create np.array((8,L,W)) of quantity a
        in each of the neighbors to a cell
        """
        a_shape = array.shape

        wgt_array = np.zeros((8, a_shape[0], a_shape[1]))
        nums = list(range(8))

        wgt_array[nums[0], :, :-1] = array[:, 1:]  # E
        wgt_array[nums[1], 1:, :-1] = array[:-1, 1:]  # NE
        wgt_array[nums[2], 1:, :] = array[:-1, :]  # N
        wgt_array[nums[3], 1:, 1:] = array[:-1, :-1]  # NW
        wgt_array[nums[4], :, 1:] = array[:, :-1]  # W
        wgt_array[nums[5], :-1, 1:] = array[1:, :-1]  # SW
        wgt_array[nums[6], :-1, :] = array[1:, :]  # S
        wgt_array[nums[7], :-1, :-1] = array[1:, 1:]  # SE

        if fix_edges:
            wgt_array[nums[0], :, -1] = wgt_array[nums[0], :, -2]
            wgt_array[nums[1], :, -1] = wgt_array[nums[1], :, -2]
            wgt_array[nums[7], :, -1] = wgt_array[nums[7], :, -2]
            wgt_array[nums[1], 0, :] = wgt_array[nums[1], 1, :]
            wgt_array[nums[2], 0, :] = wgt_array[nums[2], 1, :]
            wgt_array[nums[3], 0, :] = wgt_array[nums[3], 1, :]
            wgt_array[nums[3], :, 0] = wgt_array[nums[3], :, 1]
            wgt_array[nums[4], :, 0] = wgt_array[nums[4], :, 1]
            wgt_array[nums[5], :, 0] = wgt_array[nums[5], :, 1]
            wgt_array[nums[5], -1, :] = wgt_array[nums[5], -2, :]
            wgt_array[nums[6], -1, :] = wgt_array[nums[6], -2, :]
            wgt_array[nums[7], -1, :] = wgt_array[nums[7], -2, :]

        if normalize:
            a_sum = np.sum(wgt_array, axis=0)
            wgt_array[:, a_sum != 0] = wgt_array[
                :, a_sum != 0] / a_sum[a_sum != 0]

        return wgt_array


@njit
def _get_weight_at_cell_water(ind, weight_sfc, weight_int, depth_nbrs, ct_nbrs,
                              dry_depth, gamma, theta):
    """Compute water weights for a given cell.

    This is a jitted function called by :func:`_get_water_weight_array`.
    """
    # create a fixed set of bools
    dry = (depth_nbrs < dry_depth)
    wall = (ct_nbrs == -2)
    ctr = (np.arange(9) == 4)
    drywall = np.logical_or(dry, wall)
    invalid = np.logical_or(wall, ctr)
    if ind[0] == 0:
        invalid[:3] = True

    # set drywall to zero
    weight_sfc[drywall] = 0
    weight_int[drywall] = 0

    # set ctr to 0 before rebalancing
    weight_sfc[ctr] = 0
    weight_int[ctr] = 0

    # sanity check
    if np.any(np.isnan(weight_sfc)) or np.any(np.isnan(weight_int)):
        raise RuntimeError('NaN encountered in input to water weighting. '
                           'Please report error.')

    # initial rebalance weights
    if np.sum(weight_sfc) > 0:
        weight_sfc = weight_sfc / np.sum(weight_sfc)
    if np.sum(weight_int) > 0:
        weight_int = weight_int / np.sum(weight_int)

    weight = gamma * weight_sfc + (1 - gamma) * weight_int
    weight = (depth_nbrs ** theta) * weight

    # enforce disallowed choice to not move
    weight[ctr] = 0

    # correct the weights for random choice
    if np.sum(weight) > 0:
        # if any cells have positive weights, rebalance to sum() == 1
        weight = weight / np.sum(weight)

    elif np.sum(weight) == 0:
        # if all weights are zeros/nan
        if wall[4] == True:  # noqa: E712
            # if current cell is wall cell
            #    just set to zero and pass
            weight[:] = 0
        else:
            # convert to random walk into...
            wetvalid = np.logical_and(~dry, ~invalid)
            nwetvalid = np.sum(wetvalid)
            if nwetvalid > 0:
                # any wet (and valid)
                weight[:] = 0
                weight[wetvalid] = (1 / nwetvalid)
            else:
                # no wet (and valid) neighbors. How did we get here??
                #    Maybe by flooding corrections? Solution is to
                #    allow to walk into any non-land cell, true random
                nnotwall = np.sum(~wall)
                weight[:] = 0
                if nnotwall > 0:
                    weight[~wall] = (1 / nnotwall)
                else:
                    raise RuntimeError('No non-wall cells surrounding cell. '
                                       'Please report error.')

            weight = weight / np.sum(weight)

    else:
        raise RuntimeError('Water sum(weight) less than 0. '
                           'Please report error.')

    # final sanity check
    if np.any(np.isnan(weight)):
        raise RuntimeError('NaN encountered in return from water weighting. '
                           'Please report error.')

    return weight


# @njit('(float32[:,:], float32[:,:], int64[:,:], float32[:,:], float32[:,:],'
#       'float32[:], float32[:], float32[:],'
#       'float64, float64, float64)')
@njit
def _get_water_weight_array(depth, stage, cell_type, qx, qy,
                            ivec_flat, jvec_flat, distances_flat,
                            dry_depth, gamma, theta_water):
    """Worker for :obj:`_get_water_weight_array`.

    This is a jitted function which handles the actual computation of looping
    through locations of the model domain and determining the water
    weighting.

    See :meth:`get_water_weight_array` for more information.

    .. note::

        If you are trying to change water weighting behavior, consider
        reimplementing this method, which calls a custom version of
        :func:`_get_weight_at_cell_water`.
    """
    L, W = depth.shape
    pad_stage = shared_tools.custom_pad(stage)
    pad_depth = shared_tools.custom_pad(depth)
    pad_cell_type = shared_tools.custom_pad(cell_type)

    water_weights = np.zeros((L, W, 9))

    for i in range(L):
        for j in range(W):
            stage_nbrs = pad_stage[i - 1 + 1:i + 2 + 1, j - 1 + 1:j + 2 + 1]
            depth_nbrs = pad_depth[i - 1 + 1:i + 2 + 1, j - 1 + 1:j + 2 + 1]
            ct_nbrs = pad_cell_type[i - 1 + 1:i + 2 + 1, j - 1 + 1:j + 2 + 1]

            weight_sfc, weight_int = shared_tools.get_weight_sfc_int(
                stage[i, j], stage_nbrs.ravel(),
                qx[i, j], qy[i, j], ivec_flat, jvec_flat,
                distances_flat)

            water_weights[i, j] = _get_weight_at_cell_water(
                (i, j), weight_sfc, weight_int,
                depth_nbrs.ravel(), ct_nbrs.ravel(),
                dry_depth, gamma, theta_water)

    return water_weights


@njit('int64[:](int64[:], float64[:,:])')
def _choose_next_directions(inds, water_weights):
    """Get new cell locations, based on water weights.

    Algorithm is to:
        1. loop through each parcel, which is described by a pair in the
           `inds` array.
        2. determine the water weights for that location (from `water_weights`)
        3. choose a new cell based on the probabilities of the weights (using
           the `random_pick` function)

    Parameters
    ----------
    inds : :obj:`ndarray`
        Current unraveled indices of the parcels. ``(N,)``  `ndarray`
        containing the unraveled indices.

    water_weights : :obj:`ndarray`
        Weights of every water cell. ``(LxW, 9)`` `ndarray`, uses unraveled
        indicies along 0th dimension; 9 cells represent self and 8 neighboring
        cells.

    Returns
    -------
    next_direction : :obj:`ndarray`
        The direction to move towards the new cell for water parcels, relative
        to the current location. I.e., this is the D8 direction the parcel is
        going to travel in the next stage,
        :obj:`pyDeltaRCM.shared_tools._calculate_new_ind`.
    """
    next_direction = np.zeros_like(inds)
    for p in range(inds.shape[0]):
        ind = inds[p]
        if ind != 0:
            weight = water_weights[ind, :]
            next_direction[p] = shared_tools.random_pick(weight)
        else:
            next_direction[p] = 4

    return next_direction


@njit('int64[:](int64[:], int64[:], int64[:])')
def _calculate_new_inds(current_inds, new_direction, ravel_walk):
    """Calculate the new location (current_inds) of parcels.

    Use the information of the current parcel (`current_inds`) in conjunction
    with the D8 direction the parcel needs to travel (`new_direction`) to
    determine the new current_inds of each parcel.

    In implementation, we use the flattened `ravel_walk` array, but the result
    is identical to unraveling the index, adding `iwalk` and `jwalk` to the
    location and then raveling the index back.

    .. code::

        ind_tuple = shared_tools.custom_unravel(ind, domain_shape)
        new_ind = (ind_tuple[0] + jwalk[newd],
                   ind_tuple[1] + iwalk[newd])
        new_inds[p] = shared_tools.custom_ravel(new_ind, domain_shape)


    """
    # preallocate return array
    new_inds = np.zeros_like(current_inds)

    # loop through every parcel
    for p in range(current_inds.shape[0]):

        # extract current_ind and direction
        ind = current_inds[p]
        newd = new_direction[p]

        # check if the parcel moves
        if newd != 4:
            # if moves, compute new ind for parcel
            new_inds[p] = ind + ravel_walk[newd]
        else:
            # if not moves, set new ind to 0
            #  (should be only those already at 0)
            new_inds[p] = 0

    return new_inds


@njit
def _check_for_loops(free_surf_walk_inds, new_inds, _step,
                     L0, CTR, stage_above_SL):
    """Check for loops in water parcel pathways.

    Look for looping random walks. I.e., this function checks for where a
    parcel will return on its :obj:`new_inds` to somewhere it has already been
    in :obj:`free_surf_walk_inds`. If the loop is found, the parcel is
    relocated along the mean transport vector of the parcel, which is computed
    as the vector from the cell `(0, CTR)` to the new location in `new_inds`.

    This implementation of loop checking will relocate any parcel that has
    looped, but only disqualifies a parcel `p` from contributing to the free
    surface in :obj:`_accumulate_free_surf_walks` (i.e., `looped[p] == 1`) if
    the stage at the looped location is above the sea level in the domain.

    Parameters
    ----------
    free_surf_walk_inds
        Array recording the walk of parcels. Shape is `(:obj:`Np_water`,
        ...)`, where the second dimension will depend on the step number, but
        records each step of the parcel. Each element in the array records the
        *flat* index into the domain.

    new_inds
        Array recording the new index for each water parcel, if the step is
        taken. Shape is `(Np_water, 1)`, with each element recording
        the *flat* index into the domain shape.

    _step
        Step number of water parcels.

    L0
        Domain shape parameter, number of cells inlet length.

    CTR
        Domain shape parameter, index along inlet wall making the center of
        the domain. I.e., `(0, CTR)` is the midpoint across the inlet, along
        the inlet domain edge.

    stage_above_SL
        Water surface elevation minus the domain sea level.

    Returns
    -------
    new_inds
        An updated array of parcel indicies, where the index of a parcel has
        been changed, if and only if, that parcel was looped.

    looped
        A binary integer array indicating whether a parcel was determined to
        have been looped, and should be disqualified from the free surface
        computation.

    Examples
    --------

    The following shows an example of how water parcels that looped along
    their paths would be relocated. Note than in this example, the parcels are
    artificially forced to loop, just for the sake of demonstration.

    .. plot:: water_tools/_check_for_loops.py

    """
    nparcels = free_surf_walk_inds.shape[0]
    domain_shape = stage_above_SL.shape
    domain_min_x = domain_shape[0] - 2
    domain_min_y = domain_shape[1] - 2
    L0_ind_cut = ((L0) * domain_shape[1])-1

    looped = np.zeros_like(new_inds)

    stage_v_SL = np.abs(stage_above_SL) < 1e-1  # true if they are same

    # if the _step number is larger than the inlet length
    if (_step > L0):
        # loop though every parcel walk
        for p in np.arange(nparcels):
            new_ind = new_inds[p]  # the new index of the parcel
            full_walk = free_surf_walk_inds[p, :]  # the parcel's walk
            nonz_walk = full_walk[full_walk > 0]   # where non-zero
            relv_walk = nonz_walk[nonz_walk > L0_ind_cut]

            if (new_ind > 0):

                # determine if has a repeat ind
                has_repeat_ind = False
                for _, iind in enumerate(relv_walk):
                    if iind == new_ind:
                        has_repeat_ind = True
                        break

                if has_repeat_ind:
                    # handle when a loop is detected
                    px0, py0 = shared_tools.custom_unravel(
                        new_ind, domain_shape)

                    # compute a new location for the parcel along the
                    #   mean-transport vector
                    Fx = px0 - 1
                    Fy = py0 - CTR
                    Fw = np.sqrt(Fx**2 + Fy**2)

                    # relocate the parcel along mean-transport vector
                    if Fw != 0:
                        px = px0 + int(np.round(Fx / Fw * 5.))
                        py = py0 + int(np.round(Fy / Fw * 5.))

                    # limit the new px and py to beyond the inlet, and
                    #   away from domain edges
                    px = np.minimum(domain_min_x, np.maximum(px, L0))
                    py = np.minimum(domain_min_y, np.maximum(1, py))

                    # ravel the index for return
                    nind = shared_tools.custom_ravel((px, py), domain_shape)
                    new_inds[p] = nind

                    # only disqualify the parcel if it has not reached sea
                    #   level by the time it loops
                    if not stage_v_SL[px0, py0]:
                        looped[p] = 1  # this parcel is looped

    return new_inds, looped


@njit
def _update_dirQfield(qfield, dist, inds, astep, dirstep):
    """Update unit vector of water flux in x or y."""
    for i, ii in enumerate(inds):
        if astep[i]:
            qfield[ii] += dirstep[i] / dist[i]
    return qfield


@njit
def _update_absQfield(qfield, dist, inds, astep, Qp_water, dx):
    """Update norm of water flux vector."""
    for i, ii in enumerate(inds):
        if astep[i]:
            qfield[ii] += Qp_water / dx / 2
    return qfield


@njit
def _accumulate_free_surface_walks(free_surf_walk_inds, free_surf_flag,
                                   cell_type, uw, ux, uy, depth,
                                   dx, u0, h0, H_SL, S0):
    """Accumulate the free surface by walking parcel paths.

    This routine comprises the hydrodynamic physics-based computations.

    Algorithm is to:
        1. loop through every parcel's directed random walk in series.

        2. for a parcel's walk, unravel the indices and determine whether the
           parcel should contribute to the free surface. Parcels are
           considered contributors if they have reached the ocean and if they
           are not looped pathways.

        3. then, we begin at the downstream end of the parcel's walk and
           iterate up-walk until, determining the `Hnew` for each location.
           Downstream of the shoreline-ocean boundary, the water surface
           elevation is set to the sea level. Upstream of the shoreline-ocean
           boundary, the water surface is determined according to the
           land-slope (:obj:`S0`) and the parcel pathway.

        4. repeat from 2, for each parcel.


    Examples
    --------

    The following shows an example of the walk of a few water parcels, along
    with the resultant computed water surface.

    .. plot:: water_tools/_accumulate_free_surface_walks.py

    """
    _shape = uw.shape
    Hnew = np.zeros(_shape)
    sfc_visit = np.zeros(_shape)
    sfc_sum = np.zeros(_shape)

    # for every parcel, walk the path of the parcel
    for p, inds in enumerate(free_surf_walk_inds):

        # unravel the indices of the parcel into `xs` and `ys`
        inds_whr = inds[inds > 0]  # where the path has meaningful values
        xs = np.zeros_like(inds_whr)  # x coordinates
        ys = np.zeros_like(inds_whr)  # x coordinates
        for pp, ind_whr in np.ndenumerate(inds_whr):
            xs[pp], ys[pp] = shared_tools.custom_unravel(ind_whr, _shape)

        # determine whether the pathway contributes to the free surface
        Hnew[:] = 0
        if ((cell_type[xs[-1], ys[-1]] == -1) and (free_surf_flag[p] == 1)):
            # if cell is in ocean, H = H_SL (downstream boundary condition)
            Hnew[xs[-1], ys[-1]] = H_SL

            # counting back from last cell visited
            in_ocean = True  # whether we are in the ocean or not
            dH = 0
            for it in range(len(xs) - 2, -1, -1):
                i = xs[it]
                ip = xs[it + 1]
                j = ys[it]
                jp = ys[it + 1]

                # if the parcel has moved at all
                if (i != ip) or (j != jp):
                    # if in the ocean (not reached the shoreline yet)
                    if in_ocean:
                        # see if it is shoreline
                        if ((uw[i, j] > u0 * 0.5) or (depth[i, j] < 0.1 * h0)):
                            in_ocean = False  # passed the shoreline
                    # otherwise, in the delta
                    else:
                        # if no velocity
                        if uw[i, j] == 0:
                            dH = 0  # no change in water surface elevation
                        else:
                            # diff between streamline and parcel path
                            dH = (S0 * (ux[i, j] * (ip - i) * dx +
                                        uy[i, j] * (jp - j) * dx) / uw[i, j])

                # previous cell's surface plus difference in H
                Hnew[i, j] = Hnew[ip, jp] + dH

                # add up # of cell visits
                sfc_visit[i, j] = sfc_visit[i, j] + 1

                # sum of all water surface elevations
                sfc_sum[i, j] = sfc_sum[i, j] + Hnew[i, j]

    return sfc_visit, sfc_sum


@njit
def _smooth_free_surface(Hin, cell_type, Nsmooth, Csmooth):
    """Smooth the free surface.

    Parameters
    ----------
    Hin
        Stage input to the smoothing (i.e., the old stage).

    cell_type
        Type of each cell.

    Nsmooth
        Number of times to smooth the free surface (i.e., stage)

    Csmooth
        Underrelaxation coefficient for smoothing iterations.

    Examples
    --------
    The following shows an example of the water surface smoothing process.

    .. plot:: water_tools/_smooth_free_surface.py
    """
    # grab relevant shape information
    L, W = Hin.shape

    # pad the input stage and cell type arrays
    cell_type_pad = shared_tools.custom_pad(cell_type)

    # create copy of H which is modified in following smoothing
    Htemp = np.copy(Hin)
    for _ in range(Nsmooth):

        # create another copy to refernce as base in Nsmooth iteration
        Hsmth = np.copy(Htemp)
        Hsmth_pad = np.copy(shared_tools.custom_pad(Hsmth))

        # loop through all cells and determine a smoothed index
        for i in range(L):
            for j in range(W):

                # locate non-edge cells
                if cell_type[i, j] != -1:

                    # slice the padded array neighbors
                    cell_type_nbrs = cell_type_pad[
                        i - 1 + 1:i + 2 + 1, j - 1 + 1:j + 2 + 1]
                    Hsmth_nbrs = Hsmth_pad[
                        i - 1 + 1:i + 2 + 1, j - 1 + 1:j + 2 + 1]

                    # flatten
                    Hsmth_nbrs = Hsmth_nbrs.flatten()
                    cell_type_nbrs = cell_type_nbrs.flatten()

                    # create a validation array
                    invalid_nbrs = (cell_type_nbrs == -2)  # where land
                    invalid_nbrs[4] = True

                    # invalid cells cannot contribute stage values
                    Hsmth_nbrs[invalid_nbrs] = 0

                    # contributed stage and number of contributing nbrs
                    sumH = np.sum(Hsmth_nbrs)
                    nbcount = np.sum(~invalid_nbrs)

                    # if there are any nbrs
                    if nbcount > 0:
                        # the new stage is the underrelaxed average of nbrs
                        Htemp[i, j] = ((Csmooth * Hsmth[i, j]) +
                                       ((1 - Csmooth) * (sumH / nbcount)))

    # return the smoothed stage
    Hsmth = np.copy(Htemp)
    return Hsmth



import warnings

import numpy as np
import matplotlib.pyplot as plt

import pyDeltaRCM

# filter out the warning raised about no netcdf being found
warnings.filterwarnings("ignore", category=UserWarning)


# init delta model
with pyDeltaRCM.shared_tools._docs_temp_directory() as output_dir:
    delta = pyDeltaRCM.DeltaModel(
        out_dir=output_dir,
        resume_checkpoint='../../_resources/checkpoint')

_shp = delta.eta.shape

delta.route_all_sand_parcels()

_eta_before = np.copy(delta.eta)
delta.topo_diffusion()
_eta_after = np.copy(delta.eta)

# set up axis
fig, ax = plt.subplots(1, 3, sharex=True, sharey=True, figsize=(12, 4))

# fill in axis0
pyDeltaRCM.debug_tools.plot_domain(
    _eta_before, ax=ax[0], grid=False, cmap='cividis', vmin=-6)
ax[0].set_title('bed elevation before \n topographic diffusion')

# fill in axis1
pyDeltaRCM.debug_tools.plot_domain(
    _eta_after, ax=ax[1], grid=False, cmap='cividis', vmin=-6)
ax[1].set_title('bed elevation before \n topographic diffusion')


# fill in axis2
_diff = _eta_after - _eta_before
pyDeltaRCM.debug_tools.plot_domain(
    _diff, ax=ax[2], grid=False, cmap='RdBu',
    vmin=-0.1, vmax=0.1)
ax[2].set_title('difference (after - before)')


plt.tight_layout()
plt.show()




import os
import logging
import warnings
import platform
import sys

from math import floor
import numpy as np

from scipy.sparse import lil_matrix, csr_matrix

from netCDF4 import Dataset
import time as time_lib
import yaml
import abc

from . import shared_tools
from . import sed_tools

# tools for initiating deltaRCM model domain


class init_tools(abc.ABC):

    def init_output_infrastructure(self):
        """Initialize the output infrastructure (folder and save lists).

        This method is the first called in the initialization of the
        `DeltaModel`, after the configuration variables have been imported.
        """
        # output directory config
        self.prefix = self.out_dir
        self.prefix_abspath = os.path.abspath(self.prefix)

        # create directory if it does not exist
        if not os.path.exists(self.prefix_abspath):
            os.makedirs(self.prefix_abspath)
            assert os.path.isdir(self.prefix_abspath)  # validate dir created

        self._save_fig_list = dict()  # dict of figure variables to save
        self._save_var_list = dict()  # dict of variables to save
        self._save_var_list['meta'] = dict()  # set up meta dict

    def init_logger(self):
        """Initialize a logger.

        The logger is initialized regardless of the value of ``self.verbose``.
        The level of information printed to the log depends on the verbosity
        setting.
        """
        timestamp = time_lib.strftime('%Y%m%d-%H%M%S')
        self.logger = logging.getLogger(self.prefix_abspath + timestamp)
        self.logger.setLevel(logging.INFO)

        # create the logging file handler
        fh = logging.FileHandler(os.path.join(
            self.prefix_abspath, 'pyDeltaRCM_' + timestamp + '.log'))
        formatter = logging.Formatter(
            '%(asctime)s - %(levelname)s - %(message)s')
        fh.setFormatter(formatter)

        # add handler to logger object
        self.logger.addHandler(fh)

        _msg = 'Output log file initialized'
        self.log_info(_msg, verbosity=0)

        # log attributes of the model and environment
        self.log_info('pyDeltaRCM version {}'.format(self.__pyDeltaRCM_version__))  # log the pyDeltaRCM version
        self.log_info('Python version {}'.format(sys.version),
                      verbosity=0)  # log the python version
        self.log_info('Platform: {}'.format(platform.platform()),
                      verbosity=0)  # log the os

    def import_files(self, kwargs_dict={}):
        """Import the input files.

        This method handles the parsing and validation of any options supplied
        via the configuration.

        Parameters
        ----------
        kwargs_dict : :obj:`dict`, optional

            A dictionary with keys matching valid model parameter names that
            can be specified in a configuration YAML file. Keys given in this
            dictionary will supercede values specified in the YAML
            configuration.

        Returns
        -------
        """
        # This dictionary serves as a container to hold values for both the
        # user-specified file and the internal defaults.
        input_file_vars = dict()

        # get the special loader from the shared tools
        loader = shared_tools.custom_yaml_loader()

        # Open and access both yaml files --> put in dictionaries
        # parse default yaml and find expected types
        default_file = open(self.default_file, mode='r')
        default_dict = yaml.load(default_file, Loader=loader)
        default_file.close()
        for k, v in default_dict.items():
            if not type(v['type']) is list:
                default_dict[k]['type'] = [eval(v['type'])]
            else:
                default_dict[k]['type'] = [eval(_v) for _v in v['type']]

        # only access the user input file if provided.
        if self.input_file:
            try:
                user_file = open(self.input_file, mode='r')
                user_dict = yaml.load(user_file, Loader=loader)
                user_file.close()
            except ValueError as e:
                raise e
        else:
            user_dict = dict()

        # replace values in the user yaml file with anything specifed in the
        #   **kwargs input
        for kwk, kwv in kwargs_dict.items():
            if kwk in user_dict.keys():
                warnings.warn(UserWarning(
                    'A keyword specification was also found in the '
                    'user specified input YAML file: %s' % kwk))
            user_dict[kwk] = kwv

        # go through and populate input vars with user and default values,
        # checking user values for correct type.
        for k, v in default_dict.items():
            if k in user_dict:
                expected_type = v['type']
                if type(user_dict[k]) in expected_type:
                    input_file_vars[k] = user_dict[k]
                else:
                    raise TypeError(f'Input for {str(k)} not of the '
                                    f'right type in yaml configuration '
                                    f'file {self.input_file}. '
                                    f'Input type was {type(k).__name__}, '
                                    f'but needs to be {expected_type}.')
            else:
                input_file_vars[k] = default_dict[k]['default']

        # add custom subclass yaml parameters (yaml or defaults) to input vars
        for k, v in self.subclass_parameters.items():
            if k in input_file_vars:
                warnings.warn(UserWarning(
                    'Custom subclass parameter name is already a '
                    'default yaml parameter of the model, '
                    'custom parameter value will not be used.'
                ))
            elif k in user_dict:
                # get expected types
                if not type(v['type']) is list:
                    expected_type = [eval(v['type'])]
                else:
                    expected_type = [eval(_v) for _v in v['type']]
                # evaluate against expected types
                if type(user_dict[k]) in expected_type:
                    input_file_vars[k] = user_dict[k]
                else:
                    raise TypeError(f'Input for {str(k)} not of the '
                                    f'right type in yaml configuration '
                                    f'file {self.input_file}. '
                                    f'Input type was {type(k).__name__}, '
                                    f'but needs to be {expected_type}.')
            else:
                # set using default value
                input_file_vars[k] = v['default']

        # save the input file as a hidden attr and grab needed value
        self._input_file_vars = input_file_vars
        self.out_dir = self._input_file_vars['out_dir']
        self.verbose = self._input_file_vars['verbose']
        if self._input_file_vars['legacy_netcdf']:
            self._netcdf_coords = ('total_time', 'length', 'width')
        else:
            self._netcdf_coords = ('time', 'x', 'y')

    def process_input_to_model(self):
        """Process input file to model variables.

        Loop through the items specified in the model configuration and apply
        them to the model (i.e., ``self``). Additionally, write the input
        values specified into the log file.

        .. note::

            If ``self.resume_checkpoint == True``, then the input values are
            *not* written to the log.
        """
        _msg = 'Setting up model configuration'
        self.log_info(_msg, verbosity=0)

        _msg = f'Model type is: {self.__class__.__name__}'
        self.log_info(_msg, verbosity=0)

        # process the input file to attributes of the model
        for k, v in list(self._input_file_vars.items()):
            setattr(self, k, v)

        # write the input file values to the log
        if not self._resume_checkpoint:
            for k, v in list(self._input_file_vars.items()):
                _msg = 'Configuration variable `{var}`: {val}'.format(
                    var=k, val=v)
                self.log_info(_msg, verbosity=0)

    def determine_random_seed(self):
        """Set the random seed if given.

        If a random seed is specified, set the seed to this value.

        Writes the seed to the log for record.
        """
        if self._seed is None:
            # generate a random seed for reproducibility
            self.seed = np.random.randint((2**32) - 1, dtype='u8')

        shared_tools.set_random_seed(self._seed)

        # always write the seed to file for record and reproducability
        _msg = 'Random seed is: %s ' % str(self._seed)
        self.log_info(_msg, verbosity=0)

    def create_other_variables(self):
        """Model implementation variables.

        Creates variables for model implementation, from specified boundary
        condition variables. This method is run during initial model
        instantition. Internally, this method calls :obj:`set_constants` and
        :obj:`create_boundary_conditions`.

        .. note::

            It is usually not necessary to re-run this method if you change
            boundary condition variables, and instead only re-run
            :obj:`create_boundary_conditions`.
        """
        _msg = 'Setting other variables'
        self.log_info(_msg, verbosity=1)

        self.init_Np_water = self._Np_water
        self.init_Np_sed = self._Np_sed

        self.dx = float(self._dx)

        self.theta_sand = self._coeff_theta_sand * self._theta_water
        self.theta_mud = self._coeff_theta_mud * self._theta_water

        self.U_dep_mud = self._coeff_U_dep_mud * self._u0
        self.U_ero_sand = self._coeff_U_ero_sand * self._u0
        self.U_ero_mud = self._coeff_U_ero_mud * self._u0

        self.L = int(round(self._Length / self._dx))        # num cells in x
        self.W = int(round(self._Width / self._dx))         # num cells in y

        # cross-stream center of domain idx
        self.CTR = floor(self.W / 2.) - 1
        if self.CTR <= 1:
            self.CTR = floor(self.W / 2.)

        self.set_constants()

        self.create_boundary_conditions()

        self._save_any_grids = (self._save_eta_grids or
                                self._save_depth_grids or
                                self._save_stage_grids or
                                self._save_discharge_grids or
                                self._save_velocity_grids or
                                self._save_sedflux_grids or
                                self._save_sandfrac_grids or
                                self._save_discharge_components or
                                self._save_velocity_components)
        if self._save_any_grids:  # always save metadata if saving grids
            self._save_metadata = True

        self._is_finalized = False

    def set_constants(self):
        """Set the model constants.

        Configure constants, including coordinates and distances, as well as
        environmental constants (gravity), and kernels for smoothing
        topography.

        Some of the constants defined herein:
            * `self.g`, gravitational acceleration
            * `self.distances`, distance from cell `i,j` to neighbors (and self)  # noqa: E501
            * `self.iwalk`, step distance cross domain to cell in indexed direction
            * `self.jwalk`, step distance down domain to cell in indexed direction
            * `self.ravel_walk`, flattened index distance to cell in indexed direction

        Each of these attributes also has a `self.xxxxx_flat` sibling
        attribute, which is simply the flattened version of that attribute.
        """
        _msg = 'Setting model constants'
        self.log_info(_msg, verbosity=1)

        # simple constants
        self.g = 9.81   # (gravitation const.)
        sqrt2 = np.sqrt(2)
        sqrt05 = np.sqrt(0.5)

        # translations arrays
        self.distances = np.array([[sqrt2,    1,  sqrt2],
                                   [1,        1,      1],
                                   [sqrt2,    1,  sqrt2]], dtype=np.float32)
        self.ivec      = np.array([[-sqrt05,  0,  sqrt05],  # noqa: E221
                                   [-1,       0,       1],
                                   [-sqrt05,  0,  sqrt05]], dtype=np.float32)
        self.jvec      = np.array([[-sqrt05, -1, -sqrt05],  # noqa: E221
                                   [0,        0,       0],
                                   [sqrt05,   1,  sqrt05]], dtype=np.float32)
        self.iwalk     = np.array([[-1,       0,       1],  # noqa: E221
                                   [-1,       0,       1],
                                   [-1,       0,       1]], dtype=np.int64)
        self.jwalk     = np.array([[-1,      -1,      -1],  # noqa: E221
                                   [0,        0,       0],
                                   [1,        1,       1]], dtype=np.int64)

        # derivatives of translations
        self.distances_flat = self.distances.flatten()
        self.ivec_flat = self.ivec.flatten()
        self.jvec_flat = self.jvec.flatten()
        self.iwalk_flat = self.iwalk.flatten()
        self.jwalk_flat = self.jwalk.flatten()
        self.ravel_walk = (self.jwalk * self.W) + self.iwalk  # walk, flattened
        self.ravel_walk_flat = self.ravel_walk.flatten()

        # kernels for topographic smoothing
        self.kernel1 = np.array([[1, 1, 1],
                                 [1, -8, 1],
                                 [1, 1, 1]]).astype(np.int64)

        self.kernel2 = np.array([[1, 1, 1],
                                 [1, 0, 1],
                                 [1, 1, 1]]).astype(np.int64)

    def create_boundary_conditions(self):
        """Create model boundary conditions

        This method is run during model initialization to determine the
        boundary conditions based on the initial conditions specified as model
        input.

        However, the method also should be called when certain boundary
        condition variables are updated (e.g., during some custom model runs).
        For example, if some property of the inlet flow condition is changed,
        you will *likely* want to re-run this method (e.g., `u0`), because
        there are several other parameters that depend on the value of the
        inlet flow velocity (e.g., `Qw0` and `Qs0`); of course, your
        scientific question may choose to leave these parameters as they are
        too (in which case you should *not* re-run this method). See
        :doc:`/examples/updating_boundary_conditions` for more information.

        .. note::

            This method is automatically called for the "named" variables used
            by the BMI wrapper (e.g., `channel_flow_velocity`,
            `channel_width`, `channel_flow_depth`, and
            `influx_sediment_concentration`., so you do not need to call it
            again if you modify the model boundary conditions that way.
        """
        # inlet length and width
        self.L0 = max(
            1, min(int(round(self._L0_meters / self._dx)), self.L // 4))
        self.N0 = max(
            3, min(int(round(self._N0_meters / self._dx)), self.W // 4))

        self.u_max = 2.0 * self._u0  # maximum allowed flow velocity
        self.C0 = self._C0_percent * 1 / 100.  # sediment concentration

        # (m) critial depth to switch to "dry" node
        self.dry_depth = min(0.1, 0.1 * self._h0)

        self.gamma = (self.g * self.S0 * self._dx /
                      (self.u0**2))  # water weighting coeff

        # (m^3) reference volume, volume to fill cell to characteristic depth
        self.V0 = self.h0 * (self._dx**2)
        self.Qw0 = self.u0 * self.h0 * self.N0 * self._dx    # const discharge

        # at inlet
        self.qw0 = self.u0 * self.h0  # water unit input discharge
        self.Qp_water = self.Qw0 / self._Np_water  # volume each water parcel
        self.qs0 = self.qw0 * self.C0  # sed unit discharge
        self.dVs = 0.1 * self.N0**2 * self.V0  # total sed added per timestep
        self.Qs0 = self.Qw0 * self.C0  # sediment total input discharge
        self.Vp_sed = self.dVs / self._Np_sed  # volume of each sediment parcel

        # max number of jumps for parcel
        if self.stepmax is None:
            self.stepmax = 2 * (self.L + self.W)
        else:
            self.stepmax = int(self.stepmax)

        # initial width of self.free_surf_walk_indices
        self.size_indices = int(self.stepmax / 2)

        self._dt = self.dVs / self.Qs0  # time step size

        self.omega_flow_iter = 2. / self._itermax

        # number of times to repeat topo diffusion
        self.N_crossdiff = int(round(self.dVs / self.V0))

        self._lambda = self._sed_lag  # sedimentation lag

        self.diffusion_multiplier = (self._dt / self.N_crossdiff * self._alpha
                                     * 0.5 / self._dx**2)

    def create_domain(self):
        """Create the model domain.

        This method initializes the model domain, including coordinate arrays,
        gridded fields (eta, qw, etc.) and cell type information. The method
        is called during initialization of the `DeltaModel`, and likely does
        not need to be called again.

        .. hint::

            If you need to modify the model domain after it has been created,
            it is probably safe to modify attributes directly, but take care
            to ensure any dependent fields are also approrpriately changed.
        """
        _msg = 'Creating model domain'
        self.log_info(_msg, verbosity=1)

        # resolve any boundary conditions
        self._hb = self.hb or self.h0  # basin depth

        # ---- coordinates ----
        self.xc = np.arange(0, self.L) * self._dx
        self.yc = np.arange(0, self.W) * self._dx
        self.x, self.y = np.meshgrid(np.arange(0, self.W),
                                     np.arange(0, self.L))
        self.X, self.Y = np.meshgrid(np.arange(0, self.W+1)*self._dx,
                                     np.arange(0, self.L+1)*self._dx)

        # ---- empty arrays ----
        self.cell_type = np.zeros((self.L, self.W), dtype=np.int64)
        self.eta = np.zeros((self.L, self.W), dtype=np.float32)
        self.eta0 = np.copy(self.eta)  # establish eta0 copy
        self.stage = np.zeros((self.L, self.W), dtype=np.float32)
        self.depth = np.zeros((self.L, self.W), dtype=np.float32)
        self.qx = np.zeros((self.L, self.W), dtype=np.float32)
        self.qy = np.zeros((self.L, self.W), dtype=np.float32)
        self.qxn = np.zeros((self.L, self.W), dtype=np.float32)
        self.qyn = np.zeros((self.L, self.W), dtype=np.float32)
        self.qwn = np.zeros((self.L, self.W), dtype=np.float32)
        self.ux = np.zeros((self.L, self.W), dtype=np.float32)
        self.uy = np.zeros((self.L, self.W), dtype=np.float32)
        self.uw = np.zeros((self.L, self.W), dtype=np.float32)
        self.qs = np.zeros((self.L, self.W), dtype=np.float32)
        self.sand_frac = np.zeros((self.L, self.W), dtype=np.float32) + \
            self.sand_frac_bc
        self.active_layer = np.zeros((self.L, self.W), dtype=np.float32) + \
            self.sand_frac_bc
        self.Vp_dep_sand = np.zeros((self.L, self.W), dtype=np.float32)
        self.Vp_dep_mud = np.zeros((self.L, self.W), dtype=np.float32)

        # arrays for computing the free surface after water iteration
        self.free_surf_flag = np.zeros((self._Np_water,), dtype=np.int64)
        self.free_surf_walk_inds = np.zeros(
            (self._Np_water, self.size_indices), dtype=np.int64)
        self.sfc_visit = np.zeros_like(self.depth)
        self.sfc_sum = np.zeros_like(self.depth)

        # ---- domain ----
        cell_land = -2
        cell_channel = 1
        cell_ocean = 0
        cell_edge = -1

        self.cell_type[:self.L0, :] = cell_land

        channel_inds = int(self.CTR - round(self.N0 / 2)) + 1
        y_channel_max = channel_inds + self.N0
        self.cell_type[:self.L0, channel_inds:y_channel_max] = cell_channel

        self.stage[:] = (np.maximum(0, self.L0 - self.y - 1) *
                         self._dx * self._S0)
        self.stage[self.cell_type == cell_ocean] = 0.

        self.depth[self.cell_type == cell_ocean] = self.hb
        self.depth[self.cell_type == cell_channel] = self.h0

        self.qx[self.cell_type == cell_channel] = self.qw0
        self.qx[self.cell_type == cell_ocean] = self.qw0 / 5.
        self.qw = (self.qx**2 + self.qy**2)**(0.5)

        self.ux[self.depth > 0] = self.qx[
            self.depth > 0] / self.depth[self.depth > 0]
        self.uy[self.depth > 0] = self.qy[
            self.depth > 0] / self.depth[self.depth > 0]
        self.uw[self.depth > 0] = self.qw[
            self.depth > 0] / self.depth[self.depth > 0]

        # reset the land cell_type to -2
        self.cell_type[self.cell_type == cell_land] = -2
        self.cell_type[-1, :] = cell_edge
        self.cell_type[:, 0] = cell_edge
        self.cell_type[:, -1] = cell_edge

        bounds = [(np.sqrt((i - 3)**2 + (j - self.CTR)**2))
                  for i in range(self.L)
                  for j in range(self.W)]
        bounds = np.reshape(bounds, (self.L, self.W))

        self.cell_type[bounds >= min(self.L - 5, self.W / 2 - 5)] = cell_edge

        self.cell_type[:self.L0, :] = cell_land
        self.cell_type[:self.L0, channel_inds:y_channel_max] = cell_channel

        self.inlet = np.array(np.unique(np.where(self.cell_type == 1)[1]))
        self.eta[:] = self.stage - self.depth

    def init_sediment_routers(self):
        """Initialize the sediment router object here.

        These are preinitialized because the "boxing" for jitted functions is
        expensive, so we avoid boxing up the constants on each iteration.

        .. important::

            If you change any model boundary conditions, you likely should
            reinitialize the sediment routers (i.e., rerun this method)
        """
        _msg = 'Initializing sediment routers'
        self.log_info(_msg, verbosity=1)

        # initialize the MudRouter object
        self._mr = sed_tools.MudRouter(self._dt, self._dx, self.Vp_sed,
                                       self.u_max, self.U_dep_mud,
                                       self.U_ero_mud,
                                       self.ivec_flat, self.jvec_flat,
                                       self.iwalk_flat, self.jwalk_flat,
                                       self.distances_flat, self.dry_depth,
                                       self._lambda, self._beta,  self.stepmax,
                                       self.theta_mud)
        # initialize the SandRouter object
        self._sr = sed_tools.SandRouter(self._dt, self._dx, self.Vp_sed,
                                        self.u_max, self.qs0, self._u0,
                                        self.U_ero_sand,
                                        self._f_bedload,
                                        self.ivec_flat, self.jvec_flat,
                                        self.iwalk_flat, self.jwalk_flat,
                                        self.distances_flat, self.dry_depth,
                                        self._beta, self.stepmax,
                                        self.theta_sand)

    def init_output_file(self):
        """Creates a netCDF file to store output grids.

        Fills with default variables.

        .. warning::

            Overwrites an existing netcdf file with the same name if
            :attr:`~pyDeltaRCM.model.DeltaModel.clobber_netcdf` is `True`.

        """
        _msg = 'Initializing output NetCDF4 file'
        self.log_info(_msg, verbosity=1)

        # set standard/default metadata values in the dict structure
        if self._save_metadata:
            self.init_metadata_list()

        if (self._save_metadata or
                self._save_any_grids):

            directory = self.prefix
            filename = 'pyDeltaRCM_output.nc'

            file_path = os.path.join(directory, filename)
            _msg = 'Target output NetCDF4 file: {file}'.format(
                file=file_path)
            self.log_info(_msg, verbosity=2)

            if (os.path.exists(file_path)) and \
                    (self._clobber_netcdf is False):
                raise FileExistsError(
                    'Existing NetCDF4 output file in target output location: '
                    '{file_path}'.format(file_path=file_path))
            elif (os.path.exists(file_path)) and \
                    (self._clobber_netcdf is True):
                _msg = 'Replacing existing netCDF file'
                self.logger.warning(_msg)
                warnings.warn(UserWarning(_msg))
                os.remove(file_path)

            self.output_netcdf = Dataset(file_path, 'w',
                                         format='NETCDF4')

            self.output_netcdf.description = 'Output from pyDeltaRCM'
            self.output_netcdf.history = ('Created '
                                          + time_lib.ctime(time_lib.time()))
            self.output_netcdf.source = 'pyDeltaRCM v{ver}'.format(
                ver=self.__pyDeltaRCM_version__)

            # create master dimensions (pulls from `self._netcdf_coords`)
            self.output_netcdf.createDimension(self._netcdf_coords[1], self.L)
            self.output_netcdf.createDimension(self._netcdf_coords[2], self.W)
            self.output_netcdf.createDimension(self._netcdf_coords[0], None)

            # create master coordinates (as netCDF variables)
            time = self.output_netcdf.createVariable(
                'time', 'f4', (self._netcdf_coords[0],))
            time.units = 'second'

            if self._legacy_netcdf:
                # old format is 2d array x and y
                x = self.output_netcdf.createVariable(
                    'x', 'f4', self._netcdf_coords[1:])
                y = self.output_netcdf.createVariable(
                    'y', 'f4', self._netcdf_coords[1:])
                x[:] = self.x
                y[:] = self.y

            else:
                # new output format is 1d x and y
                x = self.output_netcdf.createVariable(
                    'x', 'f4', ('x'))
                y = self.output_netcdf.createVariable(
                    'y', 'f4', ('y'))
                x[:] = self.xc
                y[:] = self.yc

            x.units = 'meter'
            y.units = 'meter'

            # set up variables for output data grids
            def _create_grid_variable(varname, varunits,
                                      vartype='f4', vardims=()):
                _v = self.output_netcdf.createVariable(
                    varname, vartype, vardims)
                _v.units = varunits

            _var_list = list(self._save_var_list.keys())
            _var_list.remove('meta')
            for _val in _var_list:
                _create_grid_variable(_val, self._save_var_list[_val][1],
                                      self._save_var_list[_val][2],
                                      self._save_var_list[_val][3])

            # set up metadata group and populate variables
            def _create_meta_variable(varname, varvalue, varunits,
                                      vartype='f4', vardims=()):
                _v = self.output_netcdf.createVariable(
                    'meta/'+varname, vartype, vardims)
                _v.units = varunits
                _v[:] = varvalue

            self.output_netcdf.createGroup('meta')
            for _val in self._save_var_list['meta'].keys():
                # time-varying initialize w/ None value
                if (self._save_var_list['meta'][_val][0] is None):
                    _create_meta_variable(
                        _val, self._save_var_list['meta'][_val][0],
                        self._save_var_list['meta'][_val][1],
                        self._save_var_list['meta'][_val][2],
                        self._save_var_list['meta'][_val][3])
                # for scalars, get the attribute and store it
                else:
                    _create_meta_variable(
                        _val, getattr(
                            self, self._save_var_list['meta'][_val][0]),
                        self._save_var_list['meta'][_val][1],
                        self._save_var_list['meta'][_val][2],
                        self._save_var_list['meta'][_val][3])

            _msg = 'Output netCDF file created'
            self.log_info(_msg, verbosity=2)

    def init_subsidence(self):
        """Initialize subsidence pattern.

        Initializes patterns of subsidence if toggle_subsidence is True
        (default False). Default behavior is for the entire basin to subside
        at a constant rate (with the exception of the land boundary cells along
        the inlet boundary).

        To customize the subsiding region, or even vary the subsiding region
        over the course of the model run, we recommend subclassing the
        DeltaModel class.

        """
        _msg = 'Initializing subsidence'
        self.log_info(_msg, verbosity=1)

        if self._toggle_subsidence:

            self.subsidence_mask = np.ones((self.L, self.W), dtype=bool)
            self.subsidence_mask[:self.L0, :] = False

            self.sigma = self.subsidence_mask * self._subsidence_rate * self.dt

    def init_metadata_list(self):
        """Populate the list of metadata information.

        Sets up the dictionary object for the standard metadata.
        """
        # fixed metadata
        self._save_var_list['meta']['L0'] = ['L0', 'cells', 'i8', ()]
        self._save_var_list['meta']['N0'] = ['N0', 'cells', 'i8', ()]
        self._save_var_list['meta']['CTR'] = ['CTR', 'cells', 'i8', ()]
        self._save_var_list['meta']['dx'] = ['dx', 'meters', 'f4', ()]
        self._save_var_list['meta']['h0'] = ['h0', 'meters', 'f4', ()]
        self._save_var_list['meta']['hb'] = ['hb', 'meters', 'f4', ()]
        self._save_var_list['meta']['cell_type'] = ['cell_type',
                                                    'type', 'i8',
                                                    self._netcdf_coords[1:]]
        # subsidence metadata
        if self._toggle_subsidence:
            self._save_var_list['meta']['start_subsidence'] = [
                'start_subsidence', 'seconds', 'i8', ()
            ]
            self._save_var_list['meta']['sigma'] = [
                'sigma', 'meters per timestep', 'f4',
                self._netcdf_coords[1:]
            ]
        # time-varying metadata
        self._save_var_list['meta']['H_SL'] = [None, 'meters', 'f4',
                                               (self._netcdf_coords[0])]
        self._save_var_list['meta']['f_bedload'] = [None, 'fraction',
                                                    'f4',
                                                    (self._netcdf_coords[0])]
        self._save_var_list['meta']['C0_percent'] = [None, 'percent',
                                                     'f4',
                                                     (self._netcdf_coords[0])]
        self._save_var_list['meta']['u0'] = [None, 'meters per second',
                                             'f4',
                                             (self._netcdf_coords[0])]

    def load_checkpoint(self, defer_output=False):
        """Load the checkpoint from the .npz file.

        Uses the file at the path determined by `self.prefix` and a file named
        `checkpoint.npz`. There are a few pathways for loading, which depend
        on 1) the status of additional grid-saving options, 2) the presence of
        a netCDF output file at the expected output location, and 3) the status
        of the :obj:`defer_output` parameter passed to this method as an
        optional argument.

        As a standard user, you should not need to worry about any of these
        pathways or options. However, if you are developing pyDeltaRCM or
        customizing the model in any way that involves loadind from
        checkpoints, you should be aware of these pathways.

        For example, loading from checkpoint will succeed if no netCDF4 file
        is found, where one is expected (e.g., because :obj:`_save_any_grids`
        is `True`). In this case, a new output netcdf file will be created,
        after a `UserWarning` is issued.

        .. important::

            If you are customing the model and intend to use checkpointing and
            the :obj:`Preprocessor` parallel infrastructure, be sure that
            parameter :obj:`defer_output` is `True` until the
            :obj:`load_checkpoint` method can be called from the thread the
            model will execute on. Failure to do so may result in unexpected
            behavior with indexing in the output netCDF4 file.

        Parameters
        ----------
        defer_output : :obj:`bool`, optional
            Whether to defer any netCDF activities at present. Manipulating
            this variable is critical for parallel operations. See note above.
            Default is `False`.
        """
        _msg = 'Loading from checkpoint.'
        self.log_info(_msg, verbosity=0)

        _msg = 'Locating checkpoint file'
        self.log_info(_msg, verbosity=2)
        ckp_file = os.path.join(self._checkpoint_folder, 'checkpoint.npz')
        checkpoint = np.load(ckp_file, allow_pickle=True)

        # write saved variables back to the model
        _msg = 'Loading variables and grids into model'
        self.log_info(_msg, verbosity=2)

        # load time and counter vars
        self._time = float(checkpoint['time'])
        self.H_SL = float(checkpoint['H_SL'])
        self._time_iter = int(checkpoint['time_iter'])
        self._save_iter = int(checkpoint['save_iter'])
        self._save_time_since_data = int(checkpoint['save_time_since_data'])

        # load grids
        self.eta = checkpoint['eta']
        self.depth = checkpoint['depth']
        self.stage = checkpoint['stage']
        self.uw = checkpoint['uw']
        self.ux = checkpoint['ux']
        self.uy = checkpoint['uy']
        self.qw = checkpoint['qw']
        self.qx = checkpoint['qx']
        self.qy = checkpoint['qy']
        self.sand_frac = checkpoint['sand_frac']
        self.active_layer = checkpoint['active_layer']

        # load and set random state to continue as if run hadn't stopped
        _msg = 'Loading random state'
        self.log_info(_msg, verbosity=2)
        rng_state = tuple(checkpoint['rng_state'])
        shared_tools.set_random_state(rng_state)

        # handle the case with a netcdf file
        if ((self._save_any_grids or self._save_metadata)
                and not defer_output):

            # check if the file exists already
            #    if it does, it needs to be flushed of certain fields
            file_path = os.path.join(self.prefix, 'pyDeltaRCM_output.nc')
            if not os.path.isfile(file_path):
                _msg = ('NetCDF4 output file not found, but was expected. '
                        'Creating a new output file.')
                self.logger.warning(_msg)
                warnings.warn(UserWarning(_msg))

                # create a new file
                # reset output file counters
                self._save_iter = int(0)
                self.hook_init_output_file()
                self.init_output_file()
            else:
                # check if file is open in another process, if so throw error
                try:
                    _dataset = Dataset(file_path, 'r+', format='NETCDF')
                    _dataset.close()  # if this worked then close it
                except OSError:
                    raise RuntimeError(
                        'Could not open the NetCDF file for checkpointing. '
                        'This could be because the file is corrupt, or open '
                        'in another interpreter. Be sure to close any '
                        'connections to the file before proceeding.')
                # if not open elsewhere, then proceed
                # rename the old netCDF4 file
                _msg = 'Renaming old NetCDF4 output file'
                self.log_info(_msg, verbosity=2)
                _tmp_name = os.path.join(self.prefix,
                                         'old_pyDeltaRCM_output.nc')
                os.rename(file_path, _tmp_name)

                # write dims / attributes / variables to new netCDF file
                _msg = 'Creating NetCDF4 output file'
                self.log_info(_msg, verbosity=2)

                # populate default metadata list
                if self._save_metadata:
                    self.init_metadata_list()

                # copy data from old netCDF4 into new one
                with Dataset(_tmp_name) as src, Dataset(file_path, 'w',
                                                        format='NETCDF4') as dst:
                    # copy attributes
                    for name in src.ncattrs():
                        dst.setncattr(name, src.getncattr(name))
                    # copy dimensions
                    for name, dimension in src.dimensions.items():
                        if dimension.isunlimited():
                            dst.createDimension(name, None)
                        else:
                            dst.createDimension(name, len(dimension))
                    # copy groups (meta)
                    for name in src.groups.keys():
                        dst.createGroup(name)
                        for vname, variable in src.groups[name].variables.items():
                            _mname = name + '/' + vname
                            dst.createVariable(_mname, variable.datatype,
                                               variable.dimensions)
                            dst.groups[name].variables[vname][:] = \
                                src.groups[name].variables[vname][:]
                    # copy variables except ones to exclude
                    for name, variable in src.variables.items():
                        dst.createVariable(name, variable.datatype,
                                           variable.dimensions)
                        dst.variables[name][:] = src.variables[name][:]

                # set object attribute for model
                self.output_netcdf = Dataset(file_path, 'r+', format='NETCDF4')

                # synch netcdf file
                self.output_netcdf.sync()

                # delete old netCDF4 file
                os.remove(_tmp_name)

        _msg = 'Successfully loaded checkpoint.'
        self.log_info(_msg, verbosity=1)
